#!/usr/bin/python
# -*- coding: latin-1 -*-
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 3, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTIBILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.

"""M�dulo para obtener c�digo de autorizaci�n electr�nico CAE/CAEA webservice
WSMTX de AFIP (Factura Electr�nica Mercado Interno con codificaci�n de
productos) seg�n RG2904 (opci�n A con detalle) y RG2926/10 (CAE anticipado).
"""

__author__ = "Mariano Reingart <reingart@gmail.com>"
__copyright__ = "Copyright (C) 2010-2015 Mariano Reingart"
__license__ = "GPL 3.0"
__version__ = "1.13c"

import datetime
import decimal
# import os
import sys
from .utils import (
    verifica,
    inicializar_y_capturar_excepciones,
    BaseWS,
    get_install_dir
)

HOMO = False
LANZAR_EXCEPCIONES = True
WSDL = "https://fwshomo.afip.gov.ar/wsmtxca/services/MTXCAService?wsdl"


class WSMTXCA(BaseWS):
    "Interfaz para el WebService de Factura Electr�nica Mercado Interno \
        WSMTXCA"
    _public_methods_ = [
        'CrearFactura', 'EstablecerCampoFactura', 'AgregarIva', 'AgregarItem',
        'AgregarTributo', 'AgregarCmpAsoc', 'EstablecerCampoItem', 'AgregarOpcional',
        'AutorizarComprobante', 'CAESolicitar', 'AutorizarAjusteIVA',
        'SolicitarCAEA', 'ConsultarCAEA', 'ConsultarCAEAEntreFechas',
        'InformarComprobanteCAEA', 'InformarAjusteIVACAEA',
        'InformarCAEANoUtilizado', 'InformarCAEANoUtilizadoPtoVta',
        'ConsultarUltimoComprobanteAutorizado', 'CompUltimoAutorizado',
        'ConsultarPtosVtaCAEANoInformados', 'ConsultarComprobante',
        'ConsultarTiposComprobante', 'ConsultarTiposDocumento',
        'ConsultarAlicuotasIVA', 'ConsultarCondicionesIVA', 'ConsultarMonedas',
        'ConsultarUnidadesMedida', 'ConsultarTiposTributo', 
        'ConsultarTiposDatosAdicionales', 'ConsultarCotizacionMoneda', 
        'ConsultarPuntosVentaCAE', 'ConsultarPuntosVentaCAEA', 'AnalizarXml', 
        'ObtenerTagXml', 'LoadTestXML', 'SetParametros', 'SetTicketAcceso',
        'GetParametro', 'Dummy', 'Conectar', 'DebugLog', 'SetTicketAcceso'
    ]
    _public_attrs_ = [
        'Token', 'Sign', 'Cuit',
        'AppServerStatus', 'DbServerStatus', 'AuthServerStatus',
        'XmlRequest', 'XmlResponse', 'Version', 'InstallDir',
        'LanzarExcepciones', 'Resultado', 'Obs', 'Observaciones', 'ErrCode',
        'ErrMsg', 'EmisionTipo', 'Reproceso', 'Reprocesar', 'Evento',
        'CAE', 'Vencimiento', 'Evento', 'Errores', 'Traceback', 'Excepcion',
        'CAEA', 'Periodo', 'Orden', 'FchVigDesde', 'FchVigHasta', 'FchTopeInf',
        'FchProceso', 'CbteNro', 'FechaCbte', 'PuntoVenta', 'ImpTotal'
    ]

    _reg_progid_ = "WSMTXCA"
    _reg_clsid_ = "{8128E6AB-FB22-4952-8EA6-BD41C29B17CA}"

    # Variables globales para BaseWS:
    HOMO = HOMO
    WSDL = WSDL
    Version = "%s %s" % (__version__, HOMO and 'Homologaci�n' or '')
    Reprocesar = True  # recuperar automaticamente CAE emitidos
    LanzarExcepciones = LANZAR_EXCEPCIONES
    factura = None

    def inicializar(self):
        BaseWS.inicializar(self)
        self.AppServerStatus = self.DbServerStatus = \
            self.AuthServerStatus = None
        self.Resultado = self.Motivo = self.Reproceso = ''
        self.LastID = self.LastCMP = self.CAE = self.Vencimiento = ''
        self.CAEA = None
        self.Periodo = self.Orden = ""
        self.FchVigDesde = self.FchVigHasta = ""
        self.FchTopeInf = self.FchProceso = ""
        self.CbteNro = self.FechaCbte = ImpTotal = None
        print(ImpTotal)
        self.EmisionTipo = self.Evento = ''
        self.Reproceso = ''  # no implementado

    def __analizar_errores(self, ret):
        "Comprueba y extrae errores si existen en la respuesta XML"
        if 'arrayErrores' in ret:
            errores = ret['arrayErrores']
            for error in errores:
                self.Errores.append("%s: %s" % (
                    error['codigoDescripcion']['codigo'],
                    error['codigoDescripcion']['descripcion'],
                    ))
            self.ErrMsg = '\n'.join(self.Errores)

    def Dummy(self):
        "Obtener el estado de los servidores de la AFIP"
        result = self.client.dummy()
        self.AppServerStatus = result['appserver']
        self.DbServerStatus = result['dbserver']
        self.AuthServerStatus = result['authserver']
        return True

    def CrearFactura(
        self, concepto=None, tipo_doc=None, nro_doc=None, tipo_cbte=None,
        punto_vta=None, cbt_desde=None, cbt_hasta=None, imp_total=None,
        imp_tot_conc=None, imp_neto=None, imp_subtotal=None, imp_trib=None,
        imp_op_ex=None, fecha_cbte=None, fecha_venc_pago=None,
        fecha_serv_desde=None, fecha_serv_hasta=None, moneda_id=None,
        moneda_ctz=None, observaciones=None, caea=None, fch_venc_cae=None,
        **kwargs
    ):
        "Creo un objeto factura (interna)"
        # Creo una factura electronica de exportaci�n
        fact = {
            'tipo_doc': tipo_doc, 'nro_doc':  nro_doc,
            'tipo_cbte': tipo_cbte, 'punto_vta': punto_vta,
            'cbt_desde': cbt_desde, 'cbt_hasta': cbt_hasta,
            'imp_total': imp_total, 'imp_tot_conc': imp_tot_conc,
            'imp_neto': imp_neto,
            'imp_subtotal': imp_subtotal,  # 'imp_iva': imp_iva,
            'imp_trib': imp_trib, 'imp_op_ex': imp_op_ex,
            'fecha_cbte': fecha_cbte,
            'fecha_venc_pago': fecha_venc_pago,
            'moneda_id': moneda_id, 'moneda_ctz': moneda_ctz,
            'concepto': concepto,
            'observaciones': observaciones,
            'cbtes_asoc': [],
            'tributos': [],
            'iva': [],
            'detalles': [],
            'opcionales': [],
        }
        if fecha_serv_desde:
            fact['fecha_serv_desde'] = fecha_serv_desde
        if fecha_serv_hasta:
            fact['fecha_serv_hasta'] = fecha_serv_hasta
        if caea:
            fact['caea'] = caea
        if fch_venc_cae:
            fact['fch_venc_cae'] = fch_venc_cae

        self.factura = fact
        return True

    def EstablecerCampoFactura(self, campo, valor):
        if campo in self.factura or campo in (
            'fecha_serv_desde', 'fecha_serv_hasta', 'caea', 'fch_venc_cae'
        ):
            self.factura[campo] = valor
            return True
        else:
            return False

    def AgregarCmpAsoc(self, tipo=1, pto_vta=0, nro=0, cuit=None, fecha=None, **kwargs):
        "Agrego un comprobante asociado a una factura (interna)"
        cmp_asoc = {
            'tipo': tipo,
            'pto_vta': pto_vta,
            'nro': nro
        }
        if cuit is not None:
            cmp_asoc['cuit'] = cuit
        if fecha is not None:
            cmp_asoc['fecha'] = fecha
        self.factura['cbtes_asoc'].append(cmp_asoc)
        return True

    def AgregarTributo(
        self, tributo_id, desc, base_imp, alic, importe, **kwargs
    ):
        "Agrego un tributo a una factura (interna)"
        tributo = {
            'tributo_id': tributo_id,
            'desc': desc,
            'base_imp': base_imp,
            'importe': importe,
            }
        self.factura['tributos'].append(tributo)
        return True

    def AgregarIva(self, iva_id, base_imp, importe, **kwargs):
        "Agrego un tributo a una factura (interna)"
        iva = {
            'iva_id': iva_id,
            'importe': importe,
        }
        self.factura['iva'].append(iva)
        return True

    def AgregarItem(
        self, u_mtx=None, cod_mtx=None, codigo=None, ds=None, qty=None,
        umed=None, precio=None, bonif=None, iva_id=None, imp_iva=None,
        imp_subtotal=None, **kwargs
    ):
        "Agrego un item a una factura (interna)"
        # ds = unicode(ds, "latin1") # convierto a latin1
        # Nota: no se calcula neto, iva, etc (deben venir calculados!)
        umed = int(umed)
        if umed == 99:
            imp_subtotal = -abs(float(imp_subtotal))
            imp_iva = -abs(float(imp_iva))
            item = {
                'u_mtx': u_mtx,
                'cod_mtx': cod_mtx,
                'codigo': codigo,
                'ds': ds,
                'qty': qty if umed != 99 else None,
                'umed': umed,
                'precio': precio if umed != 99 else None,
                'bonif': bonif if umed != 99 else None,
                'iva_id': iva_id,
                'imp_iva': imp_iva,
                'imp_subtotal': imp_subtotal,
            }
        self.factura['detalles'].append(item)
        return True

    def EstablecerCampoItem(self, campo, valor):
        if self.factura['detalles'] and campo in self.factura['detalles'][-1]:
            self.factura['detalles'][-1][campo] = valor
            return True
        else:
            return False
    
    def AgregarOpcional(self, opcional_id=0, valor=None, valor2=None,
                        valor3=None, valor4=None, valor5=None,
                        valor6=None, **kwarg):
        "Agrego un dato adicional a una factura (interna)"
        op = {'opcional_id': opcional_id, 'valor': valor, 'valor2': valor2,
              'valor3': valor3, 'valor4': valor4, 'valor5': valor5,
              'valor6': valor6}
        self.factura['opcionales'].append(op)
        return True

    @inicializar_y_capturar_excepciones
    def AutorizarComprobante(self):
        f = self.factura
        # contruyo la estructura a convertir en XML:
        fact = {
            'codigoTipoDocumento': f['tipo_doc'],
            'numeroDocumento': f['nro_doc'],
            'codigoTipoComprobante': f['tipo_cbte'],
            'numeroPuntoVenta': f['punto_vta'],
            'numeroComprobante': f['cbt_desde'],
            'numeroComprobante': f['cbt_hasta'],
            'importeTotal': f['imp_total'],
            'importeNoGravado': f['imp_tot_conc'],
            'importeGravado': f['imp_neto'],
            'importeSubtotal': f['imp_subtotal'],  # 'imp_iva': imp_iva,
            'importeOtrosTributos': f['tributos'] and f['imp_trib'] or None,
            'importeExento': f['imp_op_ex'],
            'fechaEmision': f['fecha_cbte'],
            'codigoMoneda': f['moneda_id'],
            'cotizacionMoneda': f['moneda_ctz'],
            'codigoConcepto': f['concepto'],
            'observaciones': f['observaciones'],
            'fechaVencimientoPago': f.get('fecha_venc_pago'),
            'fechaServicioDesde': f.get('fecha_serv_desde'),
            'fechaServicioHasta': f.get('fecha_serv_hasta'),
            'arrayComprobantesAsociados': f['cbtes_asoc'] and [{
                'comprobanteAsociado': {
                    'codigoTipoComprobante': cbte_asoc['tipo'],
                    'numeroPuntoVenta': cbte_asoc['pto_vta'],
                    'numeroComprobante': cbte_asoc['nro'],
                    'cuit': cbte_asoc.get('cuit'),
                    'fechaEmision': cbte_asoc.get('fecha'),
                }
            } for cbte_asoc in f['cbtes_asoc']] or None,
            'arrayOtrosTributos': f['tributos'] and [{
                'otroTributo': {
                    'codigo': tributo['tributo_id'],
                    'descripcion': tributo['desc'],
                    'baseImponible': tributo['base_imp'],
                    'importe': tributo['importe'],
                }
            } for tributo in f['tributos']] or None,
            'arraySubtotalesIVA': f['iva'] and [{
                'subtotalIVA': {
                    'codigo': iva['iva_id'],
                    'importe': iva['importe'],
                }
            } for iva in f['iva']] or None,
            'arrayItems': f['detalles'] and [{
                'item': {
                    'unidadesMtx': it['u_mtx'],
                    'codigoMtx': it['cod_mtx'],
                    'codigo': it['codigo'],
                    'descripcion': it['ds'],
                    'cantidad': it['qty'],
                    'codigoUnidadMedida': it['umed'],
                    'precioUnitario': it['precio'],
                    'importeBonificacion': it['bonif'],
                    'codigoCondicionIVA': it['iva_id'],
                    'importeIVA': it['imp_iva'] if int(f['tipo_cbte'])
                    not in (6, 7, 8) and it['imp_iva'] is not None
                    else None,
                    'importeItem': it['imp_subtotal'],
                }
            } for it in f['detalles']] or None,
            'arrayDatosAdicionales': f['opcionales'] and [{'datoAdicional': {
                't': dato['opcional_id'],
                'c1': dato.get('valor'),
                'c2': dato.get('valor2'),
                'c3': dato.get('valor3'),
                'c4': dato.get('valor4'),
                'c5': dato.get('valor5'),
                'c6': dato.get('valor6'),
            }} for dato in f['opcionales']] or None,
        }

        ret = self.client.autorizarComprobante(
            authRequest={
                'token': self.Token,
                'sign': self.Sign,
                'cuitRepresentada': self.Cuit
            },
            comprobanteCAERequest=fact,
        )

        # Reprocesar en caso de error (recuperar CAE emitido anteriormente)
        if self.Reprocesar and ('arrayErrores' in ret):
            for error in ret['arrayErrores']:
                err_code = error['codigoDescripcion']['codigo']
                if ret['resultado'] == 'R' and err_code == 102:
                    # guardo los mensajes xml originales
                    xml_request = self.client.xml_request
                    xml_response = self.client.xml_response
                    cae = self.ConsultarComprobante(
                        f['tipo_cbte'],
                        f['punto_vta'],
                        ['cbt_desde'],
                        reproceso=True
                    )
                    if cae and self.EmisionTipo == 'CAE':
                        self.Reproceso = 'S'
                        self.Resultado = 'A'  # verificar O
                        return cae
                    self.Reproceso = 'N'
                    # reestablesco los mensajes xml originales
                    self.client.xml_request = xml_request
                    self.client.xml_response = xml_response

        self.Resultado = ret['resultado']  # u'A'
        if ret['resultado'] in ("A", "O"):
            cbteresp = ret['comprobanteResponse']
            self.FechaCbte = cbteresp['fechaEmision'].strftime("%Y/%m/%d")
            self.CbteNro = cbteresp['numeroComprobante']  # 1L
            self.PuntoVenta = cbteresp['numeroPuntoVenta']  # 4000
            # self. = cbteresp['cuit'] # 20267565393L
            # self. = cbteresp['codigoTipoComprobante']
            self.Vencimiento = cbteresp['fechaVencimientoCAE'].strftime(
                "%Y/%m/%d"
            )
            self.CAE = str(cbteresp['CAE'])  # 60423794871430L
        self.__analizar_errores(ret)

        for error in ret.get('arrayObservaciones', []):
            self.Observaciones.append("%(codigo)s: %(descripcion)s" % (
                error['codigoDescripcion'])
            )
        self.Obs = '\n'.join(self.Observaciones)

        if 'evento' in ret:
            self.Evento = '%(codigo)s: %(descripcion)s' % ret['evento']
        return self.CAE

    @inicializar_y_capturar_excepciones
    def CAESolicitar(self):
        try:
            cae = self.AutorizarComprobante() or ''
            self.Excepcion = "OK!"
        except Exception as e:
            print(e)
            cae = "ERR"
        finally:
            return cae

    @inicializar_y_capturar_excepciones
    def AutorizarAjusteIVA(self):
        """Env�a la informaci�n del comprobante de ajuste de IVA que desea
        autorizar"""
        f = self.factura
        # contruyo la estructura a convertir en XML:
        fact = {
            'codigoTipoDocumento': f['tipo_doc'],
            'numeroDocumento': f['nro_doc'],
            'codigoTipoComprobante': f['tipo_cbte'],
            'numeroPuntoVenta': f['punto_vta'],
            'numeroComprobante': f['cbt_desde'],
            'numeroComprobante': f['cbt_hasta'],
            'importeTotal': f['imp_total'],
            'importeNoGravado': f['imp_tot_conc'],
            'importeGravado': f['imp_neto'],
            'importeSubtotal': f['imp_subtotal'],  # 'imp_iva': imp_iva,
            'importeOtrosTributos': f['tributos'] and f['imp_trib'] or None,
            'importeExento': f['imp_op_ex'],
            'fechaEmision': f['fecha_cbte'],
            'codigoMoneda': f['moneda_id'],
            'cotizacionMoneda': f['moneda_ctz'],
            'codigoConcepto': f['concepto'],
            'observaciones': f['observaciones'],
            'fechaVencimientoPago': f.get('fecha_venc_pago'),
            'fechaServicioDesde': f.get('fecha_serv_desde'),
            'fechaServicioHasta': f.get('fecha_serv_hasta'),
            'arrayComprobantesAsociados': f['cbtes_asoc'] and [{
                'comprobanteAsociado': {
                    'codigoTipoComprobante': cbte_asoc['tipo'],
                    'numeroPuntoVenta': cbte_asoc['pto_vta'],
                    'numeroComprobante': cbte_asoc['nro'],
                }
            } for cbte_asoc in f['cbtes_asoc']] or None,
            'arrayOtrosTributos': f['tributos'] and [{
                'otroTributo': {
                    'codigo': tributo['tributo_id'],
                    'descripcion': tributo['desc'],
                    'baseImponible': tributo['base_imp'],
                    'importe': tributo['importe'],
                }
            } for tributo in f['tributos']] or None,
            'arraySubtotalesIVA': f['iva'] and [{
                'subtotalIVA': {
                    'codigo': iva['iva_id'],
                    'importe': iva['importe'],
                }
            } for iva in f['iva']] or None,
            'arrayItems': f['detalles'] and [{
                'item': {
                    'unidadesMtx': it['u_mtx'],
                    'codigoMtx': it['cod_mtx'],
                    'codigo': it['codigo'],
                    'descripcion': it['ds'],
                    'cantidad': it['qty'],
                    'codigoUnidadMedida': it['umed'],
                    'precioUnitario': it['precio'],
                    'importeBonificacion': it['bonif'],
                    'codigoCondicionIVA': it['iva_id'],
                    'importeIVA': it['imp_iva'] if int(f['tipo_cbte'])
                    not in (6, 7, 8) and it['imp_iva'] is not None else None,
                    'importeItem': it['imp_subtotal'],
                }
            } for it in f['detalles']] or None,
            'arrayDatosAdicionales': f['opcionales'] and [{'datoAdicional': {
                't': dato['opcional_id'],
                'c1': dato.get('valor'),
                'c2': dato.get('valor2'),
                'c3': dato.get('valor3'),
                'c4': dato.get('valor4'),
                'c5': dato.get('valor5'),
                'c6': dato.get('valor6'),
            }} for dato in f['opcionales']] or None,
        }

        ret = self.client.autorizarAjusteIVA(
            authRequest={
                'token': self.Token,
                'sign': self.Sign,
                'cuitRepresentada': self.Cuit
            },
            comprobanteCAERequest=fact,
        )

        self.Resultado = ret['resultado']  # u'A'
        if ret['resultado'] in ("A", "O"):
            cbteresp = ret['comprobanteResponse']
            self.FechaCbte = cbteresp['fechaEmision'].strftime("%Y/%m/%d")
            self.CbteNro = cbteresp['numeroComprobante']  # 1L
            self.PuntoVenta = cbteresp['numeroPuntoVenta']  # 4000
            # self. = cbteresp['cuit'] # 20267565393L
            # self. = cbteresp['codigoTipoComprobante']
            self.Vencimiento = cbteresp['fechaVencimientoCAE'].strftime(
                "%Y/%m/%d"
            )
            self.CAE = str(cbteresp['CAE'])  # 60423794871430L
        self.__analizar_errores(ret)

        for error in ret.get('arrayObservaciones', []):
            self.Observaciones.append("%(codigo)s: %(descripcion)s" % (
                error['codigoDescripcion'])
            )
        self.Obs = '\n'.join(self.Observaciones)

        if 'evento' in ret:
            self.Evento = '%(codigo)s: %(descripcion)s' % ret['evento']
        return self.CAE

    @inicializar_y_capturar_excepciones
    def SolicitarCAEA(self, periodo, orden):
        "Obtener un CAEA y su respectivo per�odo de vigencia"
        ret = self.client.solicitarCAEA(
            authRequest={
                'token': self.Token,
                'sign': self.Sign,
                'cuitRepresentada': self.Cuit
            },
            solicitudCAEA={
                'periodo': periodo,
                'orden': orden
            },
        )

        self.__analizar_errores(ret)

        if 'CAEAResponse' in ret:
            res = ret['CAEAResponse']
            self.CAEA = res['CAEA']
            self.Periodo = res['periodo']
            self.Orden = res['orden']
            self.FchVigDesde = res['fechaDesde']
            self.FchVigHasta = res['fechaHasta']
            self.FchTopeInf = res['fechaTopeInforme']
            self.FchProceso = res['fechaProceso']
        return self.CAEA and str(self.CAEA) or ''

    @inicializar_y_capturar_excepciones
    def ConsultarCAEA(self, periodo=None, orden=None, caea=None):
        "M�todo de consulta de CAEA"
        if periodo and orden:
            anio, mes = int(periodo[0:4]), int(periodo[4:6])
            if int(orden) == 1:
                dias = 1, 15
            else:
                if mes in (1, 3, 5, 7, 8, 10, 12):
                    dias = 16, 31
                elif mes in (4, 6, 9, 11):
                    dias = 16, 30
                else:
                    import calendar
                    if calendar.isleap(anio):
                        dias = 16, 29  # biciesto
                    else:
                        dias = 16, 28

            fecha_desde = "%04d-%02d-%02d" % (anio, mes, dias[0])
            fecha_hasta = "%04d-%02d-%02d" % (anio, mes, dias[1])

            caeas = self.ConsultarCAEAEntreFechas(fecha_desde, fecha_hasta)
            if caeas:
                caea = caeas[0]

        if caea:
            ret = self.client.consultarCAEA(
                authRequest={
                    'token': self.Token,
                    'sign': self.Sign,
                    'cuitRepresentada': self.Cuit
                },
                CAEA=caea,
            )

            self.__analizar_errores(ret)

            if 'CAEAResponse' in ret:
                res = ret['CAEAResponse']
                self.CAEA = res['CAEA']
                self.Periodo = res['periodo']
                self.Orden = res['orden']
                self.FchVigDesde = res['fechaDesde']
                self.FchVigHasta = res['fechaHasta']
                self.FchTopeInf = res['fechaTopeInforme']
                self.FchProceso = res['fechaProceso']
        return self.CAEA and str(self.CAEA) or ''

    @inicializar_y_capturar_excepciones
    def ConsultarCAEAEntreFechas(self, fecha_desde, fecha_hasta):
        "M�todo de consulta de CAEA"

        ret = self.client.consultarCAEAEntreFechas(
            authRequest={
                'token': self.Token,
                'sign': self.Sign,
                'cuitRepresentada': self.Cuit
            },
            fechaDesde=fecha_desde,
            fechaHasta=fecha_hasta,
        )

        self.__analizar_errores(ret)

        caeas = []
        print(caeas)
        if 'arrayCAEAResponse' in ret:
            return [
                res['CAEAResponse']['CAEA'] for res in ret['arrayCAEAResponse']
            ]
        return []

    @inicializar_y_capturar_excepciones
    def InformarComprobanteCAEA(self):
        "Env�a la informaci�n del comprobante emitido y asociado a un CAEA"
        f = self.factura
        # contruyo la estructura a convertir en XML:
        fact = {
            'codigoTipoDocumento': f['tipo_doc'],
            'numeroDocumento': f['nro_doc'],
            'codigoTipoComprobante': f['tipo_cbte'],
            'numeroPuntoVenta': f['punto_vta'],
            'numeroComprobante': f['cbt_desde'],
            'numeroComprobante': f['cbt_hasta'],
            'codigoTipoAutorizacion': 'A',
            'codigoAutorizacion': f['caea'],
            'importeTotal': f['imp_total'],
            'importeNoGravado': f['imp_tot_conc'],
            'importeGravado': f['imp_neto'],
            'importeSubtotal': f['imp_subtotal'],  # 'imp_iva': imp_iva,
            'importeOtrosTributos': f['tributos'] and f['imp_trib'] or None,
            'importeExento': f['imp_op_ex'],
            'fechaEmision': f['fecha_cbte'],
            'codigoMoneda': f['moneda_id'],
            'cotizacionMoneda': f['moneda_ctz'],
            'codigoConcepto': f['concepto'],
            'observaciones': f['observaciones'],
            'fechaVencimientoPago': f.get('fecha_venc_pago'),
            'fechaServicioDesde': f.get('fecha_serv_desde'),
            'fechaServicioHasta': f.get('fecha_serv_hasta'),
            'arrayComprobantesAsociados': f['cbtes_asoc'] and [{
                'comprobanteAsociado': {
                    'codigoTipoComprobante': cbte_asoc['tipo'],
                    'numeroPuntoVenta': cbte_asoc['pto_vta'],
                    'numeroComprobante': cbte_asoc['nro'],
                }
            } for cbte_asoc in f['cbtes_asoc']] or None,
            'arrayOtrosTributos': f['tributos'] and [{
                'otroTributo': {
                    'codigo': tributo['tributo_id'],
                    'descripcion': tributo['desc'],
                    'baseImponible': tributo['base_imp'],
                    'importe': tributo['importe'],
                }
            } for tributo in f['tributos']] or None,
            'arraySubtotalesIVA': f['iva'] and [{
                'subtotalIVA': {
                    'codigo': iva['iva_id'],
                    'importe': iva['importe'],
                }
            } for iva in f['iva']] or None,
            'arrayItems': f['detalles'] and [{
                'item': {
                    'unidadesMtx': it['u_mtx'],
                    'codigoMtx': it['cod_mtx'],
                    'codigo': it['codigo'],
                    'descripcion': it['ds'],
                    'cantidad': it['qty'],
                    'codigoUnidadMedida': it['umed'],
                    'precioUnitario': it['precio'],
                    'importeBonificacion': it['bonif'],
                    'codigoCondicionIVA': it['iva_id'],
                    'importeIVA': it['imp_iva'] if int(f['tipo_cbte'])
                    not in (6, 7, 8) and it['imp_iva'] is not None else None,
                    'importeItem': it['imp_subtotal'],
                }
            } for it in f['detalles']] or None,
        }

        # fecha de vencimiento opcional (igual al �ltimo d�a de vigencia
        # del CAEA)
        if 'fch_venc_cae' in f:
            fact['fechaVencimiento'] = f['fch_venc_cae']

        ret = self.client.informarComprobanteCAEA(
            authRequest={
                'token': self.Token,
                'sign': self.Sign,
                'cuitRepresentada': self.Cuit
            },
            comprobanteCAEARequest=fact,
        )

        # Reprocesar en caso de error (verifica datos informado anteriormente)
        if self.Reprocesar and ('arrayErrores' in ret):
            for error in ret['arrayErrores']:
                err_code = error['codigoDescripcion']['codigo']
                if ret['resultado'] == 'R' and err_code == 703:
                    # guardo los mensajes xml originales
                    xml_request = self.client.xml_request
                    xml_response = self.client.xml_response
                    cae = self.ConsultarComprobante(
                        f['tipo_cbte'],
                        f['punto_vta'],
                        f['cbt_desde'],
                        reproceso=True
                    )
                    if cae == f['caea'] and self.EmisionTipo == 'CAEA':
                        self.Reproceso = 'S'
                        self.Resultado = 'A'  # verificar O
                        return cae
                    self.Reproceso = 'N'
                    # reestablesco los mensajes xml originales
                    self.client.xml_request = xml_request
                    self.client.xml_response = xml_response

        self.Resultado = ret['resultado']  # u'A'
        self.Errores = []
        if ret['resultado'] in ("A", "O"):
            cbteresp = ret['comprobanteCAEAResponse']
            self.FchProceso = ret['fechaProceso'].strftime("%Y-%m-%d")
            self.CbteNro = cbteresp['numeroComprobante']  # 1L
            self.PuntoVenta = cbteresp['numeroPuntoVenta']  # 4000
            if 'fechaVencimientoCAE' in cbteresp:
                self.Vencimiento = cbteresp['fechaVencimientoCAE'].strftime(
                    "%Y-%m-%d"
                )
            else:
                self.Vencimiento = ""
            self.CAEA = str(cbteresp['CAEA'])  # 60423794871430L
            self.EmisionTipo = 'CAEA'
        self.__analizar_errores(ret)

        for error in ret.get('arrayObservaciones', []):
            self.Observaciones.append("%(codigo)s: %(descripcion)s" % (
                error['codigoDescripcion']))
        self.Obs = '\n'.join(self.Observaciones)

        if 'evento' in ret:
            self.Evento = '%(codigo)s: %(descripcion)s' % ret['evento']
        return self.CAEA

    @inicializar_y_capturar_excepciones
    def InformarAjusteIVACAEA(self):
        "Env�a la informaci�n del comprobante de ajuste de IVA emitidos"
        f = self.factura
        # contruyo la estructura a convertir en XML:
        fact = {
            'codigoTipoDocumento': f['tipo_doc'],
            'numeroDocumento': f['nro_doc'],
            'codigoTipoComprobante': f['tipo_cbte'],
            'numeroPuntoVenta': f['punto_vta'],
            'numeroComprobante': f['cbt_desde'],
            'numeroComprobante': f['cbt_hasta'],
            'codigoTipoAutorizacion': 'A',
            'codigoAutorizacion': f['caea'],
            'importeTotal': f['imp_total'],
            'importeNoGravado': f['imp_tot_conc'],
            'importeGravado': f['imp_neto'],
            'importeSubtotal': f['imp_subtotal'],  # 'imp_iva': imp_iva,
            'importeOtrosTributos': f['tributos'] and f['imp_trib'] or None,
            'importeExento': f['imp_op_ex'],
            'fechaEmision': f['fecha_cbte'],
            'codigoMoneda': f['moneda_id'],
            'cotizacionMoneda': f['moneda_ctz'],
            'codigoConcepto': f['concepto'],
            'observaciones': f['observaciones'],
            'fechaVencimientoPago': f.get('fecha_venc_pago'),
            'fechaServicioDesde': f.get('fecha_serv_desde'),
            'fechaServicioHasta': f.get('fecha_serv_hasta'),
            'arrayComprobantesAsociados': f['cbtes_asoc'] and [{
                'comprobanteAsociado': {
                    'codigoTipoComprobante': cbte_asoc['tipo'],
                    'numeroPuntoVenta': cbte_asoc['pto_vta'],
                    'numeroComprobante': cbte_asoc['nro'],
                }
            } for cbte_asoc in f['cbtes_asoc']] or None,
            'arrayOtrosTributos': f['tributos'] and [{
                'otroTributo': {
                    'codigo': tributo['tributo_id'],
                    'descripcion': tributo['desc'],
                    'baseImponible': tributo['base_imp'],
                    'importe': tributo['importe'],
                }
            } for tributo in f['tributos']] or None,
            'arraySubtotalesIVA': f['iva'] and [{
                'subtotalIVA': {
                    'codigo': iva['iva_id'],
                    'importe': iva['importe'],
                }
            } for iva in f['iva']] or None,
            'arrayItems': f['detalles'] and [{
                'item': {
                    'unidadesMtx': it['u_mtx'],
                    'codigoMtx': it['cod_mtx'],
                    'codigo': it['codigo'],
                    'descripcion': it['ds'],
                    'cantidad': it['qty'],
                    'codigoUnidadMedida': it['umed'],
                    'precioUnitario': it['precio'],
                    'importeBonificacion': it['bonif'],
                    'codigoCondicionIVA': it['iva_id'],
                    'importeIVA': it['imp_iva'] if int(f['tipo_cbte'])
                    not in (6, 7, 8) and it['imp_iva'] is not None else None,
                    'importeItem': it['imp_subtotal'],
                }
            } for it in f['detalles']] or None,
        }

        # fecha de vencimiento opcional (igual al �ltimo d�a de vigencia
        # del CAEA)
        if 'fch_venc_cae' in f:
            fact['fechaVencimiento'] = f['fch_venc_cae']

        ret = self.client.informarAjusteIVACAEA(
            authRequest={
                'token': self.Token,
                'sign': self.Sign,
                'cuitRepresentada': self.Cuit
            },
            comprobanteCAEARequest=fact,
        )

        self.Resultado = ret['resultado']  # u'A'
        if ret['resultado'] in ("A", "O"):
            cbteresp = ret['comprobanteCAEAResponse']
            self.FchProceso = ret['fechaProceso'].strftime("%Y-%m-%d")
            self.CbteNro = cbteresp['numeroComprobante']  # 1L
            self.PuntoVenta = cbteresp['numeroPuntoVenta']  # 4000
            # self. = cbteresp['cuit'] # 20267565393L
            # self. = cbteresp['codigoTipoComprobante']
            if 'fechaVencimientoCAE' in cbteresp:
                self.Vencimiento = cbteresp['fechaVencimientoCAE'].strftime(
                    "%Y-%m-%d"
                )
            else:
                self.Vencimiento = ""
            self.CAEA = str(cbteresp['CAEA'])  # 60423794871430L
            self.EmisionTipo = 'CAEA'
        self.__analizar_errores(ret)

        for error in ret.get('arrayObservaciones', []):
            self.Observaciones.append("%(codigo)s: %(descripcion)s" % (
                error['codigoDescripcion'])
            )
        self.Obs = '\n'.join(self.Observaciones)

        if 'evento' in ret:
            self.Evento = '%(codigo)s: %(descripcion)s' % ret['evento']
        return self.CAE

    @inicializar_y_capturar_excepciones
    def InformarCAEANoUtilizado(self, caea):
        ret = self.client.informarCAEANoUtilizado(
            authRequest={
                'token': self.Token,
                'sign': self.Sign,
                'cuitRepresentada': self.Cuit
            },
            CAEA=caea,
        )
        self.Resultado = ret['resultado']  # u'A'
        if ret['resultado'] in ("A", "O"):
            self.FchProceso = ret['fechaProceso'].strftime("%Y-%m-%d")
            self.CAEA = str(ret['CAEA'])  # 60423794871430L
            self.EmisionTipo = 'CAEA'
        self.__analizar_errores(ret)
        if 'evento' in ret:
            self.Evento = '%(codigo)s: %(descripcion)s' % ret['evento']
        return self.Resultado

    @inicializar_y_capturar_excepciones
    def InformarCAEANoUtilizadoPtoVta(self, caea, punto_vta):
        ret = self.client.informarCAEANoUtilizadoPtoVta(
            authRequest={
                'token': self.Token,
                'sign': self.Sign,
                'cuitRepresentada': self.Cuit
            },
            CAEA=caea,
            numeroPuntoVenta=punto_vta,
        )
        self.Resultado = ret['resultado']  # u'A'
        if ret['resultado'] in ("A", "O"):
            self.FchProceso = ret['fechaProceso'].strftime("%Y-%m-%d")
            self.CAEA = str(ret['CAEA'])  # 60423794871430L
            self.EmisionTipo = 'CAEA'
            self.PuntoVenta = ret['numeroPuntoVenta']  # 4000
        self.__analizar_errores(ret)
        if 'evento' in ret:
            self.Evento = '%(codigo)s: %(descripcion)s' % ret['evento']
        return self.Resultado

    @inicializar_y_capturar_excepciones
    def ConsultarUltimoComprobanteAutorizado(self, tipo_cbte, punto_vta):
        ret = self.client.consultarUltimoComprobanteAutorizado(
            authRequest={
                'token': self.Token,
                'sign': self.Sign,
                'cuitRepresentada': self.Cuit
            },
            consultaUltimoComprobanteAutorizadoRequest={
                'codigoTipoComprobante': tipo_cbte,
                'numeroPuntoVenta': punto_vta
            },
        )
        nro = ret.get('numeroComprobante')
        self.__analizar_errores(ret)
        self.CbteNro = nro
        return nro is not None and str(nro) or 0

    CompUltimoAutorizado = ConsultarUltimoComprobanteAutorizado

    @inicializar_y_capturar_excepciones
    def ConsultarComprobante(
        self, tipo_cbte, punto_vta, cbte_nro, reproceso=False
    ):
        "Recuperar los datos completos de un comprobante ya autorizado"
        ret = self.client.consultarComprobante(
            authRequest={
                'token': self.Token,
                'sign': self.Sign,
                'cuitRepresentada': self.Cuit
            },
            consultaComprobanteRequest={
                'codigoTipoComprobante': tipo_cbte,
                'numeroPuntoVenta': punto_vta,
                'numeroComprobante': cbte_nro,
            },
        )
        # diferencias si hay reproceso:
        difs = []
        # analizo el resultado:
        if 'comprobante' in ret:
                cbteresp = ret['comprobante']
                if reproceso:
                    # verifico los campos registrados coincidan con los
                    # enviados:
                    f = self.factura
                    verificaciones = {
                        'codigoTipoComprobante': f['tipo_cbte'],
                        'numeroPuntoVenta': f['punto_vta'],
                        'codigoConcepto': f['concepto'],
                        'codigoTipoDocumento': f['tipo_doc'],
                        'numeroDocumento': f['nro_doc'],
                        'numeroComprobante': f['cbt_desde'],
                        'numeroComprobante': f['cbt_hasta'],
                        'fechaEmision': f['fecha_cbte'],
                        'importeTotal': decimal.Decimal(str(f['imp_total'])),
                        'importeNoGravado': decimal.Decimal(
                            str(f['imp_tot_conc'])
                        ),
                        'importeGravado': decimal.Decimal(str(f['imp_neto'])),
                        'importeExento': decimal.Decimal(str(f['imp_op_ex'])),
                        'importeOtrosTributos': f['tributos'] and
                        decimal.Decimal(str(f['imp_trib'])) or None,
                        'importeSubtotal': f['imp_subtotal'],
                        'fechaServicioDesde': f.get('fecha_serv_desde'),
                        'fechaServicioHasta': f.get('fecha_serv_hasta'),
                        'fechaVencimientoPago': f.get('fecha_venc_pago'),
                        'codigoMoneda': f['moneda_id'],
                        'cotizacionMoneda': str(
                            decimal.Decimal(str(f['moneda_ctz']))
                        ),
                        'arrayItems': [
                            {'item': {
                                'unidadesMtx': it['u_mtx'],
                                'codigoMtx': it['cod_mtx'],
                                'codigo': it['codigo'],
                                'descripcion': it['ds'],
                                'cantidad': it['qty'] and decimal.Decimal(
                                    str(it['qty'])
                                ),
                                'codigoUnidadMedida': it['umed'],
                                'precioUnitario': it['precio'] is not None
                                and decimal.Decimal(str(it['precio'])) or None,
                                # 'importeBonificacion': it['bonif'],
                                'codigoCondicionIVA': decimal.Decimal(
                                    str(it['iva_id'])
                                ),
                                'importeIVA': decimal.Decimal(
                                    str(it['imp_iva'])
                                ) if int(f['tipo_cbte']) not in (6, 7, 8) and
                                it['imp_iva'] is not None else None,
                                'importeItem': decimal.Decimal(
                                    str(it['imp_subtotal'])
                                ),
                            }}
                            for it in f['detalles']],
                        'arrayComprobantesAsociados': [
                            {'comprobanteAsociado': {
                                'codigoTipoComprobante': cbte_asoc['tipo'],
                                'numeroPuntoVenta': cbte_asoc['pto_vta'],
                                'numeroComprobante': cbte_asoc['nro']
                            }}
                            for cbte_asoc in f['cbtes_asoc']],
                        'arrayOtrosTributos': [
                            {'otroTributo': {
                                'codigo': tributo['tributo_id'],
                                'descripcion': tributo['desc'],
                                'baseImponible': decimal.Decimal(
                                    str(tributo['base_imp'])
                                ),
                                'importe': decimal.Decimal(
                                    str(tributo['importe'])
                                ),
                            }}
                            for tributo in f['tributos']],
                        'arraySubtotalesIVA': [
                            {'subtotalIVA': {
                                'codigo': iva['iva_id'],
                                'importe': decimal.Decimal(
                                    str(iva['importe'])
                                ),
                            }}
                            for iva in f['iva']],
                        }
                    verifica(verificaciones, cbteresp, difs)
                    if difs:
                        print("Diferencias:", difs)
                        self.log("Diferencias: %s" % difs)
                self.FechaCbte = cbteresp['fechaEmision'].strftime("%Y/%m/%d")
                self.CbteNro = cbteresp['numeroComprobante']  # 1L
                self.PuntoVenta = cbteresp['numeroPuntoVenta']  # 4000
                self.Vencimiento = cbteresp['fechaVencimiento'].strftime(
                    "%Y/%m/%d"
                )
                self.ImpTotal = str(cbteresp['importeTotal'])
                # 60423794871430L
                self.CAE = str(cbteresp['codigoAutorizacion'])
                self.EmisionTipo = cbteresp['codigoTipoAutorizacion'] == 'A' \
                    and 'CAEA' or 'CAE'
        self.__analizar_errores(ret)
        if not difs:
            return self.CAE

    @inicializar_y_capturar_excepciones
    def ConsultarTiposComprobante(self):
        """Este m�todo permite consultar los tipos de comprobantes habilitados
        en este WS"""
        ret = self.client.consultarTiposComprobante(
            authRequest={
                'token': self.Token,
                'sign': self.Sign,
                'cuitRepresentada': self.Cuit
            },
        )
        return ["%(codigo)s: %(descripcion)s" % p['codigoDescripcion']
                for p in ret['arrayTiposComprobante']]

    @inicializar_y_capturar_excepciones
    def ConsultarTiposDocumento(self):
        ret = self.client.consultarTiposDocumento(
            authRequest={
                'token': self.Token,
                'sign': self.Sign,
                'cuitRepresentada': self.Cuit
            },
        )
        return ["%(codigo)s: %(descripcion)s" % p['codigoDescripcion']
                for p in ret['arrayTiposDocumento']]

    @inicializar_y_capturar_excepciones
    def ConsultarAlicuotasIVA(self):
        """Este m�todo permite consultar los tipos de comprobantes habilitados
        en este WS"""
        ret = self.client.consultarAlicuotasIVA(
            authRequest={
                'token': self.Token,
                'sign': self.Sign,
                'cuitRepresentada': self.Cuit
            },
        )
        return ["%(codigo)s: %(descripcion)s" % p['codigoDescripcion']
                for p in ret['arrayAlicuotasIVA']]

    @inicializar_y_capturar_excepciones
    def ConsultarCondicionesIVA(self):
        """Este m�todo permite consultar los tipos de comprobantes habilitados
        en este WS"""
        ret = self.client.consultarCondicionesIVA(
            authRequest={
                'token': self.Token,
                'sign': self.Sign,
                'cuitRepresentada': self.Cuit
            },
        )
        return ["%(codigo)s: %(descripcion)s" % p['codigoDescripcion']
                for p in ret['arrayCondicionesIVA']]

    @inicializar_y_capturar_excepciones
    def ConsultarMonedas(self):
        """Este m�todo permite consultar los tipos de comprobantes habilitados
        en este WS"""
        ret = self.client.consultarMonedas(
            authRequest={
                'token': self.Token,
                'sign': self.Sign,
                'cuitRepresentada': self.Cuit
            },
        )
        return ["%(codigo)s: %(descripcion)s" % p['codigoDescripcion']
                for p in ret['arrayMonedas']]

    @inicializar_y_capturar_excepciones
    def ConsultarUnidadesMedida(self):
        """Este m�todo permite consultar los tipos de comprobantes habilitados
        en este WS"""
        ret = self.client.consultarUnidadesMedida(
            authRequest={
                'token': self.Token,
                'sign': self.Sign,
                'cuitRepresentada': self.Cuit
            },
        )
        return ["%(codigo)s: %(descripcion)s" % p['codigoDescripcion']
                for p in ret['arrayUnidadesMedida']]

    @inicializar_y_capturar_excepciones
    def ConsultarTiposTributo(self):
        """Este m�todo permite consultar los tipos de comprobantes habilitados
        en este WS"""
        ret = self.client.consultarTiposTributo(
            authRequest={
                'token': self.Token,
                'sign': self.Sign,
                'cuitRepresentada': self.Cuit
            },
        )
        return ["%(codigo)s: %(descripcion)s" % p['codigoDescripcion']
                for p in ret['arrayTiposTributo']]
    
    @inicializar_y_capturar_excepciones
    def ConsultarTiposDatosAdicionales(self):
        "Este método permite consultar los tipos de datos adicionales."
        ret = self.client.consultarTiposDatosAdicionales(
            authRequest={'token': self.Token, 'sign': self.Sign,
                         'cuitRepresentada': self.Cuit},
        )
        return ["%(codigo)s: %(descripcion)s" % p['codigoDescripcion']
                for p in ret['arrayTiposDatosAdicionales']]

    @inicializar_y_capturar_excepciones
    def ConsultarCotizacionMoneda(self, moneda_id):
        """Este m�todo permite consultar los tipos de comprobantes habilitados
        en este WS"""
        ret = self.client.consultarCotizacionMoneda(
            authRequest={
                'token': self.Token,
                'sign': self.Sign,
                'cuitRepresentada': self.Cuit
            },
            codigoMoneda=moneda_id,
        )
        self.__analizar_errores(ret)
        if 'cotizacionMoneda' in ret:
            return str(ret['cotizacionMoneda'])

    @inicializar_y_capturar_excepciones
    def ConsultarPuntosVentaCAE(
        self,
        fmt="%(numeroPuntoVenta)s: bloqueado=%(bloqueado)s baja=%(fechaBaja)s"
    ):
        """Este m�todo permite consultar los puntos de venta habilitados para
        CAE en este WS"""
        res = self.client.consultarPuntosVentaCAE(
            authRequest={
                'token': self.Token,
                'sign': self.Sign,
                'cuitRepresentada': self.Cuit
            },
        )
        ret = []
        for p in res['arrayPuntosVenta']:
            p = p['puntoVenta']
            if 'fechaBaja' not in p:
                p['fechaBaja'] = ""
            ret.append(fmt % p if fmt else p)
        return ret

    @inicializar_y_capturar_excepciones
    def ConsultarPuntosVentaCAEA(
        self,
        fmt="%(numeroPuntoVenta)s: bloqueado=%(bloqueado)s baja=%(fechaBaja)s"
    ):
        """Este m�todo permite consultar los puntos de venta habilitados para
        CAEA en este WS"""
        res = self.client.consultarPuntosVentaCAEA(
            authRequest={
                'token': self.Token,
                'sign': self.Sign,
                'cuitRepresentada': self.Cuit
            },
        )
        ret = []
        for p in res['arrayPuntosVenta']:
            p = p['puntoVenta']
            if 'fechaBaja' not in p:
                p['fechaBaja'] = ""
            ret.append(fmt % p if fmt else p)
        return ret

    @inicializar_y_capturar_excepciones
    def ConsultarPtosVtaCAEANoInformados(self, caea):
        """Este m�todo permite  consultar que puntos de venta a�n no fueron
        informados para  un  CAEA determinado."""
        ret = self.client.consultarPtosVtaCAEANoInformados(
            authRequest={
                'token': self.Token,
                'sign': self.Sign,
                'cuitRepresentada': self.Cuit
            },
            CAEA=caea,
        )
        return [" ".join(
            [("%s=%s" % (k, v))for k, v in list(p['puntoVenta'].items())]
        )
            for p in ret['arrayPuntosVenta']
        ]


def main():
    "Funci�n principal de pruebas (obtener CAE)"
    import os
    import time

    DEBUG = '--debug' in sys.argv

    # obteniendo el TA para pruebas
    from .wsaa import WSAA
    ta = WSAA().Autenticar("wsmtxca", "reingart.crt", "reingart.key")

    wsmtxca = WSMTXCA()
    wsmtxca.SetTicketAcceso(ta)
    wsmtxca.Cuit = "20267565393"

    cache = ""
    if "--prod" in sys.argv:
        wsdl = "https://serviciosjava.afip.gov.ar/wsmtxca/services/"
        "MTXCAService?wsdl"
    else:
        wsdl = WSDL
    wsmtxca.Conectar(cache, wsdl)

    if "--dummy" in sys.argv:
        print(wsmtxca.client.help("dummy"))
        wsmtxca.Dummy()
        print("AppServerStatus", wsmtxca.AppServerStatus)
        print("DbServerStatus", wsmtxca.DbServerStatus)
        print("AuthServerStatus", wsmtxca.AuthServerStatus)

    if "--prueba" in sys.argv:
        try:
            tipo_cbte = 1
            punto_vta = 4000
            cbte_nro = wsmtxca.ConsultarUltimoComprobanteAutorizado(
                tipo_cbte, punto_vta
            )
            fecha = datetime.datetime.now().strftime("%Y-%m-%d")
            concepto = 3
            tipo_doc = 80
            nro_doc = "30000000007"
            cbte_nro = int(cbte_nro) + 1
            cbt_desde = cbte_nro
            cbt_hasta = cbt_desde
            imp_total = "122.00"
            imp_tot_conc = "0.00"
            imp_neto = "100.00"
            imp_trib = "1.00"
            imp_op_ex = "0.00"
            imp_subtotal = "100.00"
            fecha_cbte = fecha
            fecha_venc_pago = fecha
            # Fechas del per�odo del servicio facturado (solo si concepto = 1?)
            fecha_serv_desde = fecha
            fecha_serv_hasta = fecha
            moneda_id = 'PES'
            moneda_ctz = '1.000'
            obs = "Observaciones Comerciales, libre"
            if '--caea' in sys.argv:
                periodo = fecha.replace("-", "")[:6]
                orden = 1 if fecha[-2:] < 16 else 2
                caea = wsmtxca.ConsultarCAEA(periodo, orden)
            else:
                caea = None

            wsmtxca.CrearFactura(
                concepto, tipo_doc, nro_doc, tipo_cbte, punto_vta,
                cbt_desde, cbt_hasta, imp_total, imp_tot_conc, imp_neto,
                imp_subtotal, imp_trib, imp_op_ex, fecha_cbte, fecha_venc_pago,
                fecha_serv_desde, fecha_serv_hasta, moneda_id, moneda_ctz,
                obs, caea
            )

            # tipo = 19
            # pto_vta = 2
            # nro = 1234
            # wsmtxca.AgregarCmpAsoc(tipo, pto_vta, nro)

            tributo_id = 99
            desc = 'Impuesto Municipal Matanza'
            base_imp = "100.00"
            alic = "1.00"
            importe = "1.00"
            wsmtxca.AgregarTributo(tributo_id, desc, base_imp, alic, importe)

            iva_id = 5  # 21%
            base_imp = 100
            importe = 21
            wsmtxca.AgregarIva(iva_id, base_imp, importe)

            u_mtx = 123456
            cod_mtx = 1234567890123
            codigo = "P0001"
            ds = "Descripcion del producto P0001"
            qty = 2.00
            umed = 7
            precio = 100.00
            bonif = 0.00
            iva_id = 5
            imp_iva = 42.00
            imp_subtotal = 242.00
            wsmtxca.AgregarItem(
                u_mtx, cod_mtx, codigo, ds, qty, umed, precio, bonif, iva_id,
                imp_iva, imp_subtotal
            )

            if "--caea" not in sys.argv:
                # ejemplo descuento (sin precio unitario)
                wsmtxca.AgregarItem(
                    None, None, None, 'bonificacion', None, 99, None, None,
                    5, -21, -121
                )
                # ejemplo item solo descripci�n:
                # wsmtxca.AgregarItem(u_mtx, cod_mtx, codigo, ds, 1, umed,
                #                    0, 0, iva_id, 0, 0)
                
            # datos de Factura de Crédito Electrónica MiPyMEs (FCE):
            if '--fce' in sys.argv:
                wsmtxca.AgregarOpcional(21, "2850590940090418135201")  # CBU

            print(wsmtxca.factura)

            if '--caea' in sys.argv:
                wsmtxca.InformarComprobanteCAEA()
            else:
                wsmtxca.AutorizarComprobante()

            print("Resultado", wsmtxca.Resultado)
            print("CAE", wsmtxca.CAE)
            print("Vencimiento", wsmtxca.Vencimiento)
            print("Reproceso", wsmtxca.Reproceso)

            print(wsmtxca.Excepcion)
            print(wsmtxca.ErrMsg)

            cae = wsmtxca.CAE

            if cae:

                wsmtxca.ConsultarComprobante(tipo_cbte, punto_vta, cbte_nro)
                print("CAE consulta", wsmtxca.CAE, wsmtxca.CAE == cae)
                print(
                    "NRO consulta", wsmtxca.CbteNro,
                    wsmtxca.CbteNro == cbte_nro
                )
                print(
                    "TOTAL consulta",
                    wsmtxca.ImpTotal,
                    wsmtxca.ImpTotal == imp_total
                )

                wsmtxca.AnalizarXml("XmlResponse")
                assert wsmtxca.ObtenerTagXml(
                    'codigoAutorizacion'
                ) == str(wsmtxca.CAE)
                assert wsmtxca.ObtenerTagXml('codigoConcepto') == str(concepto)
                assert wsmtxca.ObtenerTagXml(
                    'arrayItems', 0, 'item', 'unidadesMtx'
                ) == '123456'

        except Exception as e:
            print(e)
            print(wsmtxca.XmlRequest)
            print(wsmtxca.XmlResponse)
            print(wsmtxca.ErrCode)
            print(wsmtxca.ErrMsg)

    if "--ajustar" in sys.argv:
        try:
            tipo_cbte = 2
            punto_vta = 4000
            cbte_nro = wsmtxca.ConsultarUltimoComprobanteAutorizado(
                tipo_cbte, punto_vta
            )
            fecha = datetime.datetime.now().strftime("%Y-%m-%d")
            concepto = 3
            tipo_doc = 80
            nro_doc = "30000000007"
            cbte_nro = int(cbte_nro) + 1
            cbt_desde = cbte_nro
            cbt_hasta = cbt_desde
            imp_total = "21.00"
            imp_tot_conc = "0.00"
            imp_neto = None
            imp_trib = "0.00"
            imp_op_ex = "0.00"
            imp_subtotal = "0.00"
            fecha_cbte = fecha
            fecha_venc_pago = fecha
            # Fechas del per�odo del servicio facturado (solo si concepto = 1?)
            fecha_serv_desde = fecha
            fecha_serv_hasta = fecha
            moneda_id = 'PES'
            moneda_ctz = '1.000'
            obs = "Observaciones Comerciales, libre"
            caea = "24163778394093"
            fch_venc_cae = None

            wsmtxca.CrearFactura(
                concepto, tipo_doc, nro_doc, tipo_cbte, punto_vta,
                cbt_desde, cbt_hasta, imp_total, imp_tot_conc, imp_neto,
                imp_subtotal, imp_trib, imp_op_ex, fecha_cbte, fecha_venc_pago,
                fecha_serv_desde, fecha_serv_hasta, moneda_id, moneda_ctz,
                obs, caea, fch_venc_cae
            )

            iva_id = 5  # 21%
            base_imp = 100
            importe = 21
            wsmtxca.AgregarIva(iva_id, base_imp, importe)

            u_mtx = 1
            cod_mtx = 7790001001139
            codigo = None
            ds = "Descripcion del producto P0001"
            qty = None
            umed = 7
            precio = None
            bonif = None
            iva_id = 5
            imp_iva = 21.00
            imp_subtotal = 21.00
            wsmtxca.AgregarItem(
                u_mtx, cod_mtx, codigo, ds, qty, umed, precio, bonif, iva_id,
                imp_iva, imp_subtotal
            )

            print(wsmtxca.factura)

            if not caea:
                wsmtxca.AutorizarAjusteIVA()
            else:
                wsmtxca.InformarAjusteIVACAEA()

            print("Resultado", wsmtxca.Resultado)
            print("CAE", wsmtxca.CAE)
            print("Vencimiento", wsmtxca.Vencimiento)

            print(wsmtxca.Excepcion)
            print(wsmtxca.ErrMsg)

        except Exception as e:
            print(e)
            print(wsmtxca.XmlRequest)
            print(wsmtxca.XmlResponse)
            print(wsmtxca.ErrCode)
            print(wsmtxca.ErrMsg)
            raise

    if "--parametros" in sys.argv:
        print(wsmtxca.ConsultarTiposComprobante())
        print(wsmtxca.ConsultarTiposDocumento())
        print(wsmtxca.ConsultarAlicuotasIVA())
        print(wsmtxca.ConsultarCondicionesIVA())
        print(wsmtxca.ConsultarMonedas())
        print(wsmtxca.ConsultarUnidadesMedida())
        print(wsmtxca.ConsultarTiposTributo())

    if "--cotizacion" in sys.argv:
        print(wsmtxca.ConsultarCotizacionMoneda('DOL'))

    if "--solicitar-caea" in sys.argv:
        periodo = sys.argv[sys.argv.index("--solicitar-caea")+1]
        orden = sys.argv[sys.argv.index("--solicitar-caea")+2]

        if DEBUG:
            print(
                "Consultando CAEA para periodo %s orden %s" % (periodo, orden)
            )

        caea = wsmtxca.ConsultarCAEA(periodo, orden)
        if not caea:
            print(
                "Solicitando CAEA para periodo %s orden %s" % (periodo, orden)
            )
            caea = wsmtxca.SolicitarCAEA(periodo, orden)

        print("CAEA:", caea)

        if wsmtxca.Errores:
            print("Errores:")
            for error in wsmtxca.Errores:
                print(error)

        if DEBUG:
            print("periodo:", wsmtxca.Periodo)
            print("orden:", wsmtxca.Orden)
            print("fch_vig_desde:", wsmtxca.FchVigDesde)
            print("fch_vig_hasta:", wsmtxca.FchVigHasta)
            print("fch_tope_inf:", wsmtxca.FchTopeInf)
            print("fch_proceso:", wsmtxca.FchProceso)


# busco el directorio de instalaci�n (global para que no cambie si usan otra
# dll)
INSTALL_DIR = WSMTXCA.InstallDir = get_install_dir()


if __name__ == '__main__':

    if "--register" in sys.argv or "--unregister" in sys.argv:
        import win32com.server.register
        win32com.server.register.UseCommandLine(WSMTXCA)
    else:
        main()



#!/usr/bin/python
# -*- coding: latin-1 -*-
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 3, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTIBILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.

"""M�dulo de ejemplo para obtener c�digo de autorizaci�n de impresi�n o
electr�nico del web service WSFE de AFIP (factura electr�nica)
Basado en wsfe-client.php de Gerardo Fisanotti - DvSHyS/DiOPIN/AFIP - 08-jun-07
Definir WSDL, WSFEURL, TA
Devuelve CAE (C�digo de autorizaci�n de impresi�n o electr�nico de WSFE)
"""

__author__ = "Marcelo Alaniz (alanizmarcelo@gmail.com)"
__copyright__ = "Copyright (C) 2008 Marcelo Alaniz"
__license__ = "LGPL 3.0"
__version__ = "1.0"

import sys
from php import date, SimpleXMLElement, SoapClient

WSDL = 'wsfe.wsdl'
WSFEURL = "https://wswhomo.afip.gov.ar/wsfe/service.asmx"
#WSFEURL = "https://servicios1.afip.gov.ar/wsfe/service.asmx"
SOAP_ACTION = 'http://ar.gov.afip.dif.facturaelectronica/' # Revisar WSDL
SOAP_NS = "http://ar.gov.afip.dif.facturaelectronica/"     # Revisar WSDL

# IMPORTANTE: este servicio web es obsoleto por disposiciones de AFIP
#             y en general s�lo es mantenido por cuestiones hist�ricas
#             desde el 1/7/2011 ser�a obligatorio usar WSFEv1 (wsfev1.py)
#             (seg�n AFIP RG2845 v.1, v.1.1 v.2, v.2.1) para m�s info ver:
#             http://www.sistemasagiles.com.ar/trac/wiki/ProyectoWSFEv1

TA = 'TA.xml'
CUIT = long(20139999999)


class WSFEError(RuntimeError):
    "Clase para informar errores del WSFE"
    def __init__(self, rerror):
        self.code = str(rerror.percode)
        self.msg = unicode(rerror.perrmsg)

def recuperar_qty(client, token, sign, cuit):
    "Recuperador de cantidad m�xima de registros de detalle"
    results = client.FERecuperaQTYRequest(
        argAuth= {"Token": token, "Sign": sign, "cuit":long(cuit)}
    )
    if int(results.FERecuperaQTYRequestResult.RError.percode) != 0:
        print("Percode: %s" % results.FERecuperaQTYRequestResult.RError.percode)
        print("MSGerror: %s" % results.FERecuperaQTYRequestResult.RError.perrmsg)
        sys.exit(3)
    return int(results.FERecuperaQTYRequestResult.qty.value)

def ultnro(client, token, sign, cuit):
    "Recuperador de �ltimo n�mero de transacci�n"
    results = client.FEUltNroRequest(
        argAuth=
        {
            'Token': token,
            'Sign' : sign,
            'cuit' : cuit
        }
    )

    if int(results.FEUltNroRequestResult.RError.percode) != 0:
        raise WSFEError(results.RError)
    return int(results.FEUltNroRequestResult.nro.value)

def recuperar_last_cmp(client, token, sign, cuit, ptovta, tipocbte):
    "Recuperador de �ltimo n�mero de comprobante"
    results = client.FERecuperaLastCMPRequest(
            argAuth =
            {
                'Token': token,
                'Sign' : sign,
                'cuit' : cuit
            },
            argTCMP=
            {
                'PtoVta' : ptovta,
                'TipoCbte' : tipocbte
            }
    )
    if int(results.FERecuperaLastCMPRequestResult.RError.percode) != 0:
        raise WSFEError(results.RError)
    return int(results.FERecuperaLastCMPRequestResult.cbte_nro)

def aut(client, token, sign, cuit, id, tipo_doc=80, nro_doc=23111111113,
    tipo_cbte=1, punto_vta=1, cbt_desde=0, cbt_hasta=0,
    imp_total=0, imp_neto=0, impto_liq=0.0,
    imp_tot_conc=0, impto_liq_rni=0.00, imp_op_ex=0.00,
    fecha_cbte=None, fecha_venc_pago=None,
    presta_serv = 0, fecha_serv_desde=None, fecha_serv_hasta=None, **kwargs ):
    "Facturador (Solicitud de Autorizaci�n de Factura Electr�nica)"

    if fecha_cbte is None: fecha_cbte = date('Ymd')
    if fecha_venc_pago is None: fecha_venc_pago = date('Ymd')

    detalle = {
        'tipo_doc': tipo_doc,
        'nro_doc':  nro_doc,
        'tipo_cbte': tipo_cbte,
        'punto_vta': punto_vta,
        'cbt_desde': cbt_desde,
        'cbt_hasta': cbt_hasta,
        'imp_total': imp_total,
        'imp_tot_conc': imp_tot_conc,
        'imp_neto': imp_neto,
        'impto_liq': impto_liq,
        'impto_liq_rni': impto_liq_rni,
        'imp_op_ex': imp_op_ex,
        'fecha_cbte': fecha_cbte,
        'fecha_venc_pago': fecha_venc_pago
    }

    if fecha_serv_desde and fecha_serv_hasta:
        detalle['fecha_serv_desde'] = fecha_serv_desde
        detalle['fecha_serv_hasta'] = fecha_serv_hasta

    results = client.FEAutRequest(
            argAuth={'Token': token, 'Sign': sign, 'cuit': long(cuit)},
            Fer={
                'Fecr': {'id': id, 'cantidadreg': 1, 'presta_serv': presta_serv},
                'Fedr': {'FEDetalleRequest': detalle}
            }
        )

    if int(results.FEAutRequestResult.RError.percode) != 0:
        raise WSFEError(results.RError)

    # Resultado: Aceptado o Rechazado
    resultado = str(results.FEAutRequestResult.resultado)
    # Motivo general/del detalle:
    motivo = str(results.FEAutRequestResult.motivo)
    if motivo in ("NULL", '00'):
        motivo = str(results.FEAutRequestResult.FedResp.FEDetalleResponse.motivo)
    reproceso = str(results.FEAutRequestResult.reproceso)
    cae = str(results.FEAutRequestResult.FedResp.FEDetalleResponse.cae)
    fecha_vto = str(results.FEAutRequestResult.FedResp.FEDetalleResponse.fecha_vto)
    #fecha_vto = "%s/%s/%s" % (vto[6:8], vto[4:6], vto[0:4])

    return dict(cae=cae, fecha_vto=fecha_vto, resultado=resultado,
                motivo=motivo, reproceso=reproceso)

def dummy(client):
    results = client.FEDummy()
    print("appserver status %s" % results.appserver)
    print("dbserver status %s" % results.dbserver)
    print("authserver status %s" % results.authserver)
    #    if is_soap_fault(results):
    # print("Fault: %s" % results.faultcode)
    # print("FaultString: %s " % results.faultstring)
    #        sys.exit(1)
    return


def main():
    # cliente soap del web service
    client = SoapClient(WSFEURL,
        action = SOAP_ACTION,
        namespace = SOAP_NS,
        trace = True)

    # obteniendo el TA
    ta_string = open(TA).read()
    ta = SimpleXMLElement(ta_string)
    token = str(ta.credentials.token)
    sign = str(ta.credentials.sign)
    # fin TA

    CUIT = sys.argv[1]
    punto_vta = 2
    tipo_cbte = 1 # 1: A 6: B

    dummy(client)

    qty = recuperar_qty(client, token, sign, CUIT)
    last_id = ultnro(client, token, sign, CUIT)
    last_cbte = recuperar_last_cmp(client, token, sign, CUIT, punto_vta, tipo_cbte)
    #import pdb; pdb.set_trace()
    dic = aut(client, token, sign,
        CUIT, id=last_id +1, cbt_desde=last_cbte +1, cbt_hasta=last_cbte +1,
        tipo_doc=80, nro_doc=23111111113,
        punto_vta=punto_vta, tipo_cbte=tipo_cbte,
        imp_total=0.01, imp_neto=0.01, impto_liq=0.00)

    cae = dic['cae']

    print("QTY: %s" % qty)
    print("Last CBTE: %s" % last_cbte)
    print("Last Id: %s" % last_id)
    print("CAE: %s" % cae)


if __name__ == '__main__':
    main()



#!/usr/bin/python
# -*- coding: latin-1 -*-
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 3, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTIBILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.

"""M�dulo para acceder a web services de la afip"""
__author__ = "Mariano Reingart (mariano@gmail.com)"
__copyright__ = "Copyright (C) 2008-2015 Mariano Reingart"
__license__ = "GPL 3.0"



#!/usr/bin/python
# -*- coding: latin-1 -*-
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 3, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTIBILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.

"M�dulo para Consulta de Operaciones Cambiarias"

__author__ = "Mariano Reingart <reingart@gmail.com>"
__copyright__ = "Copyright (C) 2011 Mariano Reingart"
__license__ = "GPL 3.0"
__version__ = "1.09a"

import os
import socket
import sys
import traceback
import py3simplesoap.client
from py3simplesoap.client import SoapClient, SoapFault, parse_proxy, \
                                set_http_wrapper
from py3simplesoap.simplexml import SimpleXMLElement
from cStringIO import StringIO

HOMO = True

if HOMO:
    WSDL = "https://fwshomo.afip.gov.ar/wscoc/COCService"
    LOCATION = "https://fwshomo.afip.gob.ar:443/wscoc2/COCService"
else:
    WSDL = "https://serviciosjava.afip.gob.ar/wscoc2/COCService"
    LOCATION = "https://serviciosjava.afip.gob.ar:443/wscoc2/COCService"

def inicializar_y_capturar_excepciones(func):
    "Decorador para inicializar y capturar errores"
    def capturar_errores_wrapper(self, *args, **kwargs):
        try:
            # inicializo (limpio variables)
            self.Errores = []
            self.ErroresFormato = []
            self.Traceback = self.Excepcion = ""
            self.ErrCode = ""
            self.ErrMsg = ""
            self.CodigoSolicitud = self.FechaSolicitud = None
            self.COC = self.FechaEmisionCOC = self.CodigoDestino = None
            self.EstadoSolicitud = self.FechaEstado = None
            self.CUITComprador = self.DenominacionComprador = None
            self.CodigoMoneda = self.CotizacionMoneda = self.MontoPesos = None
            self.CUITRepresentante = self.DenominacionRepresentante = None
            self.DJAI = self.CodigoExcepcionDJAI = self.EstadoDJAI = None
            self.DJAS = self.CodigoExcepcionDJAS = self.EstadoDJAS = None
            self.MontoFOB = self.CodigoMoneda = None
            # iniciaalizo estructuras internas persistentes
            self.__detalles_solicitudes = []
            self.__detalles_cuit = []

            # llamo a la funci�n (sin reintentos)
            return func(self, *args, **kwargs)

        except SoapFault as e:
            # guardo destalle de la excepci�n SOAP
            self.ErrCode = unicode(e.faultcode)
            self.ErrMsg = unicode(e.faultstring)
            self.Excepcion = u"%s: %s" % (e.faultcode, e.faultstring, )
            if self.LanzarExcepciones:
                raise
        except Exception as e:
            ex = traceback.format_exception(sys.exc_type, sys.exc_value,
                                            sys.exc_traceback)
            self.Traceback = ''.join(ex)
            self.Excepcion = u"%s" % (e)
            if self.LanzarExcepciones:
                raise
        finally:
            # guardo datos de depuraci�n
            if self.client:
                self.XmlRequest = self.client.xml_request
                self.XmlResponse = self.client.xml_response
    return capturar_errores_wrapper


class WSCOC:
    "Interfaz para el WebService de Consulta de Operaciones Cambiarias"
    _public_methods_ = ['GenerarSolicitudCompraDivisa',
                        'GenerarSolicitudCompraDivisaTurExt',
                        'InformarSolicitudCompraDivisa',
                        'ConsultarCUIT',
                        'ConsultarCOC',
                        'AnularCOC',
                        'ConsultarSolicitudCompraDivisa',
                        'ConsultarSolicitudesCompraDivisas',
                        'ConsultarDestinosCompra',
                        'ConsultarTiposReferencia',
                        'ConsultarMonedas',
                        'ConsultarTiposDocumento',
                        'ConsultarTiposEstadoSolicitud',
                        'ConsultarMotivosExcepcionDJAI',
                        'ConsultarDestinosCompraDJAI',
                        'ConsultarMotivosExcepcionDJAS',
                        'ConsultarDestinosCompraDJAS',
                        'ConsultarDestinosCompraTipoReferencia',
                        'LeerSolicitudConsultada', 'LeerCUITConsultado',
                        'ConsultarDJAI', 'ConsultarDJAS', 'ConsultarReferencia',
                        'LeerError', 'LeerErrorFormato', 'LeerInconsistencia',
                        'LoadTestXML',
                        'AnalizarXml', 'ObtenerTagXml',
                        'Dummy', 'Conectar', 'DebugLog']
    _public_attrs_ = ['Token', 'Sign', 'Cuit',
        'AppServerStatus', 'DbServerStatus', 'AuthServerStatus',
        'XmlRequest', 'XmlResponse', 'Version', 'InstallDir',
        'Resultado', 'Inconsistencias', 'ErrCode', 'ErrMsg',
        'CodigoSolicitud', 'FechaSolicitud', 'EstadoSolicitud', 'FechaEstado',
        'COC', 'FechaEmisionCOC', 'CodigoDestino',
        'CUITComprador', 'DenominacionComprador',
        'CodigoMoneda', 'CotizacionMoneda', 'MontoPesos',
        'CUITRepresentante', 'DenominacionRepresentante',
        'TipoDoc', 'NumeroDoc', 'CUITConsultada', 'DenominacionConsultada',
		'DJAI', 'CodigoExcepcionDJAI', 'DJAS', 'CodigoExcepcionDJAS',
        'MontoFOB', 'EstadoDJAI', 'EstadoDJAS', 'Estado', 'Tipo', 'Codigo',
        'ErroresFormato', 'Errores', 'Traceback', 'Excepcion', 'LanzarExcepciones',
        ]

    _reg_progid_ = "WSCOC"
    _reg_clsid_ = "{B30406CE-326A-46D9-B807-B7916E3F1B96}"

    Version = "%s %s %s" % (__version__, HOMO and 'Homologaci�n' or '', py3simplesoap.client.__file__)
    LanzarExcepciones = False

    def __init__(self):
        self.Token = self.Sign = self.Cuit = None
        self.AppServerStatus = None
        self.DbServerStatus = None
        self.AuthServerStatus = None
        self.XmlRequest = ''
        self.XmlResponse = ''
        self.Resultado = self.Motivo = self.Reproceso = ''
        self.__analizar_solicitud({})
        self.__analizar_inconsistencias({})
        self.__analizar_errores({})
        self.__detalles_solicitudes = None
        self.__detalles_cuit = None
        self.client = None
        self.ErrCode = self.ErrMsg = self.Traceback = self.Excepcion = ""
        self.EmisionTipo = ''
        self.Reprocesar = self.Reproceso = ''  # no implementado
        self.Log = None
        self.InstallDir = INSTALL_DIR

    def __analizar_errores(self, ret):
        "Comprueba y extrae errores si existen en la respuesta XML"
        self.Errores = []
        self.ErroresFormato = []
        if 'arrayErrores' in ret:
            errores = ret['arrayErrores']
            for error in errores:
                self.Errores.append("%s: %s" % (
                    error['codigoDescripcion']['codigo'],
                    error['codigoDescripcion']['descripcion'],
                    ))
        if 'arrayErroresFormato' in ret:
            errores = ret['arrayErroresFormato']
            for error in errores:
                self.ErroresFormato.append("%s: %s" % (
                    error['codigoDescripcionString']['codigo'],
                    error['codigoDescripcionString']['descripcion'],
                    ))

    def __analizar_solicitud(self, det):
        "Analiza y extrae los datos de una solicitud"
        self.CodigoSolicitud = det.get("codigoSolicitud")
        self.FechaSolicitud = det.get("fechaSolicitud")
        self.COC = str(det.get("coc"))
        self.FechaEmisionCOC = det.get("fechaEmisionCOC")
        self.EstadoSolicitud = det.get("estadoSolicitud")
        self.FechaEstado = det.get("fechaEstado")
        self.CUITComprador = str(det.get("detalleCUITComprador",
                                     {}).get("cuit", ""))
        self.DenominacionComprador = det.get("detalleCUITComprador",
                                             {}).get("denominacion")
        self.CodigoMoneda = det.get("codigoMoneda")
        self.CotizacionMoneda = det.get("cotizacionMoneda")
        self.MontoPesos = det.get("montoPesos")
        self.CUITRepresentante = str(det.get("DetalleCUITRepresentante",
                                         {}).get("cuit", ""))
        self.DenominacionRepresentante = det.get("DetalleCUITRepresentante",
                                                 {}).get("denominacion")
        self.CodigoDestino = det.get("codigoDestino")
        self.DJAI = det.get("djai")
        self.CodigoExcepcionDJAI = det.get("codigoExcepcionDJAI")
        self.DJAS = det.get("djas")
        self.CodigoExcepcionDJAS = det.get("codigoExcepcionDJAS")
        ref = det.get("referencia")
        if ref:
            self.Tipo = ref['tipo']
            self.Codigo = ref['codigo']

    def __analizar_inconsistencias(self, ret):
        "Comprueba y extrae (formatea) las inconsistencias"
        self.Inconsistencias = []
        if 'arrayInconsistencias' in ret:
            inconsistencias = ret['arrayInconsistencias']
            for inconsistencia in inconsistencias:
                self.Inconsistencias.append("%s: %s" % (
                    inconsistencia['codigoDescripcion']['codigo'],
                    inconsistencia['codigoDescripcion']['descripcion'],
                    ))

    def __log(self, msg):
        if not isinstance(msg, unicode):
            msg = unicode(msg, 'utf8', 'ignore')
        if not self.Log:
            self.Log = StringIO()
        self.Log.write(msg)
        self.Log.write('\n\r')

    def DebugLog(self):
        "Devolver y limpiar la bit�cora de depuraci�n"
        if self.Log:
            msg = self.Log.getvalue()
            # limpiar log
            self.Log.close()
            self.Log = None
        else:
            msg = u''
        return msg

    @inicializar_y_capturar_excepciones
    def Conectar(self, cache=None, wsdl=None, proxy="", wrapper=None, cacert=None, timeout=30):
        # cliente soap del web service
        if timeout:
            self.__log("Estableciendo timeout=%s" % (timeout, ))
            socket.setdefaulttimeout(timeout)
        if wrapper:
            Http = set_http_wrapper(wrapper)
            self.Version = WSCOC.Version + " " + Http._wrapper_version
        proxy_dict = parse_proxy(proxy)
        location = LOCATION
        if HOMO or not wsdl:
            wsdl = WSDL
        elif not wsdl.endswith("?wsdl") and wsdl.startswith("http"):
            location = wsdl
            wsdl += "?wsdl"
        elif wsdl.endswith("?wsdl"):
            location = wsdl[:-5]
        if not cache or HOMO:
            # use 'cache' from installation base directory
            cache = os.path.join(self.InstallDir, 'cache')
        self.__log("Conectando a wsdl=%s cache=%s proxy=%s" % (wsdl, cache, proxy_dict))
        self.client = SoapClient(
            wsdl = wsdl,
            cache = cache,
            proxy = proxy_dict,
            ns="coc",
            cacert=cacert,
            soap_ns="soapenv",
            soap_server="jbossas6",
            trace = "--trace" in sys.argv)
        # corrijo ubicaci�n del servidor (http en el WSDL)
        self.client.services['COCService']['ports']['COCServiceHttpSoap11Endpoint']['location'] = location
        self.__log("Corrigiendo location=%s" % (location, ))
        return True

    @inicializar_y_capturar_excepciones
    def Dummy(self):
        "Obtener el estado de los servidores de la AFIP"
        result = self.client.dummy()
        ret = result['dummyReturn']
        self.AppServerStatus = ret['appserver']
        self.DbServerStatus = ret['dbserver']
        self.AuthServerStatus = ret['authserver']
        return True

    @inicializar_y_capturar_excepciones
    def GenerarSolicitudCompraDivisa(self, cuit_comprador, codigo_moneda,
                                     cotizacion_moneda, monto_pesos,
                                    cuit_representante, codigo_destino,
									djai=None, codigo_excepcion_djai=None,
                                    djas=None, codigo_excepcion_djas=None,
                                    tipo=None, codigo=None,
                                    ):
        "Generar una Solicitud de operaci�n cambiaria"
        if tipo and codigo:
            referencia = {'tipo': tipo, 'codigo': codigo}
        else:
            referencia = None
        res = self.client.generarSolicitudCompraDivisa(
            authRequest={'token': self.Token, 'sign': self.Sign, 'cuitRepresentada': self.Cuit},
            cuitComprador=cuit_comprador,
            codigoMoneda=codigo_moneda,
            cotizacionMoneda=cotizacion_moneda,
            montoPesos=monto_pesos,
            cuitRepresentante=cuit_representante,
            codigoDestino=codigo_destino,
			djai=djai, codigoExcepcionDJAI=codigo_excepcion_djai,
            djas=djas, codigoExcepcionDJAS=codigo_excepcion_djas,
            referencia=referencia,
            )

        self.Resultado = ""
        ret = res.get('generarSolicitudCompraDivisaReturn', {})
        self.Resultado = ret.get('resultado')
        det = ret.get('detalleSolicitud', {})
        self.__analizar_solicitud(det)
        self.__analizar_inconsistencias(det)
        self.__analizar_errores(ret)

        return True

    @inicializar_y_capturar_excepciones
    def GenerarSolicitudCompraDivisaTurExt(self, tipo_doc, numero_doc, apellido_nombre,
                                    codigo_moneda, cotizacion_moneda, monto_pesos,
                                    cuit_representante, codigo_destino,
                                    ):
        "Generar una Solicitud de operaci�n cambiaria"
        res = self.client.generarSolicitudCompraDivisaTurExt(
            authRequest={'token': self.Token, 'sign': self.Sign, 'cuitRepresentada': self.Cuit},
            detalleTurExtComprador={
                'tipoNumeroDoc': {'tipoDoc': tipo_doc, 'numeroDoc': numero_doc},
                'apellidoNombre': apellido_nombre},
            codigoMoneda=codigo_moneda,
            cotizacionMoneda=cotizacion_moneda,
            montoPesos=monto_pesos,
            cuitRepresentante=cuit_representante,
            codigoDestino=codigo_destino,
            )

        self.Resultado = ""
        ret = res.get('generarSolicitudCompraDivisaTurExtReturn', {})
        self.Resultado = ret.get('resultado')
        det = ret.get('detalleSolicitud', {})
        self.__analizar_solicitud(det)
        self.__analizar_inconsistencias(det)
        self.__analizar_errores(ret)

        return True

    @inicializar_y_capturar_excepciones
    def InformarSolicitudCompraDivisa(self, codigo_solicitud, nuevo_estado):
        "Informar la aceptaci�n o desistir una solicitud generada con anterioridad"

        res = self.client.informarSolicitudCompraDivisa(
            authRequest={'token': self.Token, 'sign': self.Sign, 'cuitRepresentada': self.Cuit},
            codigoSolicitud=codigo_solicitud,
            nuevoEstado=nuevo_estado,
        )

        self.Resultado = ""
        ret = res.get('informarSolicitudCompraDivisaReturn', {})
        self.Resultado = ret.get('resultado')
        self.__analizar_solicitud(ret)
        self.__analizar_errores(ret)
        return True


    @inicializar_y_capturar_excepciones
    def ConsultarCUIT(self, numero_doc, tipo_doc=80, sep="|"):
        "Consultar la CUIT, CDI � CUIL, seg�n corresponda, para un determinado tipo y n�mero de documento."

        res = self.client.consultarCUIT(
            authRequest={'token': self.Token, 'sign': self.Sign, 'cuitRepresentada': self.Cuit},
            tipoNumeroDoc={'tipoDoc': tipo_doc, 'numeroDoc': numero_doc}
        )

        self.__detalles_cuit = []

        if 'consultarCUITReturn' in res:
            ret = res['consultarCUITReturn']
            self.__analizar_errores(ret)
            if 'tipoNumeroDoc' in ret:
                self.TipoDoc = ret['tipoNumeroDoc']['tipoDoc']
                self.NumeroDoc = ret['tipoNumeroDoc']['numeroDoc']
                for detalle in ret.get('arrayDetallesCUIT', []):
                    # agrego el detalle para consultarlo luego (LeerCUITConsultado)
                    det = detalle['detalleCUIT']
                    self.__detalles_cuit.append(det)
                # devuelvo una lista de cuit/denominaci�n
                return [(u"%(cuit)s\t%(denominacion)s" %
                            d['detalleCUIT']).replace("\t", sep)
                        for d in ret.get('arrayDetallesCUIT', [])]
            else:
                return []
        else:
            self.TipoDoc = None
            self.NumeroDoc = None
            return [""]

    def LeerCUITConsultado(self):
        "Recorro los CUIT devueltos por ConsultarCUIT"

        if self.__detalles_cuit:
            # extraigo el primer item
            det = self.__detalles_cuit.pop(0)
            self.CUITConsultada = str(det['cuit'])
            self.DenominacionConsultada = str(det['denominacion'])
            return True
        else:
            return False

    @inicializar_y_capturar_excepciones
    def ConsultarCOC(self, coc):
        "Obtener los datos de un COC existente"

        res = self.client.consultarCOC(
            authRequest={'token': self.Token, 'sign': self.Sign, 'cuitRepresentada': self.Cuit},
            coc=coc,
        )
        ret = res.get('consultarCOCReturn', {})
        det = ret.get('detalleSolicitud', {})
        self.__analizar_solicitud(det)
        self.__analizar_inconsistencias(det)
        self.__analizar_errores(ret)
        return True

    @inicializar_y_capturar_excepciones
    def AnularCOC(self, coc, cuit_comprador):
        "Anular COC existente (estado CO 24hs)"
        res = self.client.anularCOC(
            authRequest={'token': self.Token, 'sign': self.Sign, 'cuitRepresentada': self.Cuit},
            coc=coc,
            cuitComprador=cuit_comprador,
            )

        ret = res.get('anularCOCReturn', {})
        self.__analizar_solicitud(ret)
        self.__analizar_inconsistencias(ret)
        self.__analizar_errores(ret)

        return True


    @inicializar_y_capturar_excepciones
    def ConsultarSolicitudCompraDivisa(self, codigo_solicitud):
        "Consultar una Solicitud de Operaci�n Cambiaria"
        res = self.client.consultarSolicitudCompraDivisa(
            authRequest={'token': self.Token, 'sign': self.Sign, 'cuitRepresentada': self.Cuit},
            codigoSolicitud=codigo_solicitud,
            )

        ret = res.get('consultarSolicitudCompraDivisaReturn', {})
        det = ret.get('detalleSolicitud', {})
        self.__analizar_solicitud(det)
        self.__analizar_errores(ret)

        return True

    @inicializar_y_capturar_excepciones
    def ConsultarSolicitudesCompraDivisas(self, cuit_comprador,
                                          estado_solicitud,
                                          fecha_emision_desde,
                                          fecha_emision_hasta,
                                        ):
        "Consultar Solicitudes de operaciones cambiarias"
        res = self.client.consultarSolicitudesCompraDivisas(
            authRequest={'token': self.Token, 'sign': self.Sign, 'cuitRepresentada': self.Cuit},
            cuitComprador=cuit_comprador,
            estadoSolicitud=estado_solicitud,
            fechaEmisionDesde=fecha_emision_desde,
            fechaEmisionHasta=fecha_emision_hasta,
            )

        self.__analizar_errores(res)

        ret = res.get('consultarSolicitudesCompraDivisasReturn', {})
        solicitudes = []                    # c�digos a devolver
        self.__detalles_solicitudes = []    # diccionario para recorrerlo luego
        for array in ret.get('arrayDetallesSolicitudes', []):
            det = array['detalleSolicitudes']
            # guardo el detalle para procesarlo luego (LeerSolicitudConsultada)
            self.__detalles_solicitudes.append(det)
            # devuelvo solo el c�digo de solicitud
            solicitudes.append(det.get("codigoSolicitud"))
        return solicitudes


    def LeerSolicitudConsultada(self):
        "Proceso de a una solicitud los detalles devueltos por ConsultarSolicitudesCompraDivisas"
        if self.__detalles_solicitudes:
            # extraigo el primer item
            det = self.__detalles_solicitudes.pop(0)
            self.__analizar_solicitud(det)
            self.__analizar_errores(det)
            self.__analizar_inconsistencias(det)
            return True
        else:
            return False

    @inicializar_y_capturar_excepciones
    def ConsultarDJAI(self, djai, cuit):
        "Consultar Declaraci�n Jurada Anticipada de Importaci�n"
        res = self.client.consultarDJAI(
            authRequest={'token': self.Token, 'sign': self.Sign, 'cuitRepresentada': self.Cuit},
            djai=djai, cuit=cuit,
            )

        ret = res.get('consultarDJAIReturn', {})
        self.__analizar_errores(ret)
        self.DJAI = ret.get('djai')
        self.MontoFOB = ret.get('montoFOB')
        self.CodigoMoneda = ret.get('codigoMoneda')
        self.EstadoDJAI = ret.get('estadoDJAI')
        return True

    @inicializar_y_capturar_excepciones
    def ConsultarDJAS(self, djas, cuit):
        "Consultar Declaraci�n Jurada Anticipada de Servicios"
        res = self.client.consultarDJAS(
            authRequest={'token': self.Token, 'sign': self.Sign, 'cuitRepresentada': self.Cuit},
            djas=djas, cuit=cuit,
            )

        ret = res.get('consultarDJASReturn', {})
        self.__analizar_errores(ret)
        self.DJAS = ret.get('djas')
        self.MontoFOB = ret.get('montoFOB')
        self.CodigoMoneda = ret.get('codigoMoneda')
        self.EstadoDJAS = ret.get('estadoDJAS')
        return True

    @inicializar_y_capturar_excepciones
    def ConsultarReferencia(self, tipo, codigo):
        "Consultar una determinada referencia seg�n su tipo (1: DJAI, 2: DJAS, 3: DJAT)"
        res = self.client.consultarReferencia(
            authRequest={'token': self.Token, 'sign': self.Sign, 'cuitRepresentada': self.Cuit},
            referencia={'tipo': tipo, 'codigo': codigo},
            )

        ret = res.get('consultarReferenciaReturn', {})
        self.__analizar_errores(ret)
        #self.Codigo = ret.get('codigo')
        self.MontoFOB = ret.get('monto')
        self.CodigoMoneda = ret.get('codigoMoneda')
        self.Estado = ret.get('estado')
        return True

    @inicializar_y_capturar_excepciones
    def ConsultarMonedas(self, sep="|"):
        "Este m�todo retorna el universo de Monedas disponibles en el presente WS, indicando c�digo y descripci�n de cada una"
        res = self.client.consultarMonedas(
            authRequest={'token': self.Token, 'sign': self.Sign, 'cuitRepresentada': self.Cuit},
            )
        ret = res['consultarMonedasReturn']
        return [("%(codigo)s\t%(descripcion)s"
                    % p['codigoDescripcion']).replace("\t", sep)
                 for p in ret['arrayMonedas']]

    @inicializar_y_capturar_excepciones
    def ConsultarDestinosCompra(self, sep="|"):
        "Consultar Tipos de Destinos de compra de divisas"
        res = self.client.consultarDestinosCompra(
            authRequest={'token': self.Token, 'sign': self.Sign, 'cuitRepresentada': self.Cuit},
            )
        #<consultarDestinosCompraReturn>22.02 m
        #<arrayDestinos>
        #<destinos>
        #<tipoDestino>TipoDestinoSimpleType</tipoDestino>
        #<arrayCodigosDescripciones>
        #<codigoDescripcion>
        #<codigo>short</codigo>
        #<descripcion>string</descripcion>
        #</codigoDescripcion>
        #</arrayCodigosDescripciones>
        #</destinos>
        #</arrayDestinos>
        ret = res['consultarDestinosCompraReturn']
        dest = []
        for array in ret['arrayDestinos']:
            destino = array['destinos']
            codigos = [("%s\t%s\t%s"
                    % (destino['tipoDestino'],
                       p['codigoDescripcion']['codigo'],
                       p['codigoDescripcion']['descripcion'],
                    )).replace("\t", sep)
                 for p in destino['arrayCodigosDescripciones']]
            dest.extend(codigos)
        return dest

    @inicializar_y_capturar_excepciones
    def ConsultarTiposDocumento(self, sep="|"):
        "Consultar Tipos de Documentos"
        res = self.client.consultarTiposDocumento(
            authRequest={'token': self.Token, 'sign': self.Sign, 'cuitRepresentada': self.Cuit},
            )
        ret = res['consultarTiposDocumentoReturn']
        return [("%(codigo)s\t%(descripcion)s"
                    % p['codigoDescripcion']).replace("\t", sep)
                 for p in ret['arrayTiposDocumento']]

    @inicializar_y_capturar_excepciones
    def ConsultarTiposEstadoSolicitud(self, sep="|"):
        "Este m�todo devuelve los diferentes tipos de estado que puede tener una solicitud."
        res = self.client.consultarTiposEstadoSolicitud(
            authRequest={'token': self.Token, 'sign': self.Sign, 'cuitRepresentada': self.Cuit},
            )
        ret = res['consultarTiposEstadoSolicitudReturn']
        return [("%(codigo)s\t%(descripcion)s"
                    % p['codigoDescripcionString']).replace("\t", sep)
                 for p in ret['arrayTiposEstadoSolicitud']]

    @inicializar_y_capturar_excepciones
    def ConsultarMotivosExcepcionDJAI(self, sep='|'):
        "Este m�todo retorna el universo de motivos de excepciones a la Declaraci�n Jurada Anticipada de Importaci�n"
        res = self.client.consultarMotivosExcepcionDJAI(
            authRequest={'token': self.Token, 'sign': self.Sign, 'cuitRepresentada': self.Cuit},
            )
        ret = res['consultarMotivosExcepcionDJAIReturn']
        return [("%(codigo)s\t%(descripcion)s"
                    % p['codigoDescripcion']).replace("\t", sep)
                 for p in ret['arrayMotivosExcepcion']]

    @inicializar_y_capturar_excepciones
    def ConsultarDestinosCompraDJAI(self, sep='|'):
        "Este m�todo retorna el subconjunto de los destinos de compra de divisas alcanzados por las normativas de la Declaraci�n Jurada Anticipada de Importaci�n."
        res = self.client.consultarDestinosCompraDJAI(
            authRequest={'token': self.Token, 'sign': self.Sign, 'cuitRepresentada': self.Cuit},
            )
        ret = res['consultarDestinosCompraDJAIReturn']
        return [("%(codigo)s\t%(descripcion)s"
                    % p['codigoDescripcion']).replace("\t", sep)
                 for p in ret['arrayCodigosDescripciones']]

    @inicializar_y_capturar_excepciones
    def ConsultarMotivosExcepcionDJAS(self, sep='|'):
        "Este m�todo retorna el universo de motivos de excepciones a la Declaraci�n Jurada Anticipada de Servicios"
        res = self.client.consultarMotivosExcepcionDJAS(
            authRequest={'token': self.Token, 'sign': self.Sign, 'cuitRepresentada': self.Cuit},
            )
        ret = res['consultarMotivosExcepcionDJASReturn']
        return [("%(codigo)s\t%(descripcion)s"
                    % p['codigoDescripcion']).replace("\t", sep)
                 for p in ret['arrayMotivosExcepcion']]

    @inicializar_y_capturar_excepciones
    def ConsultarDestinosCompraDJAS(self, sep='|'):
        "Este m�todo retorna el subconjunto de los destinos de compra de divisas alcanzados por las normativas de la Declaraci�n Jurada Anticipada de Servicios"
        res = self.client.consultarDestinosCompraDJAS(
            authRequest={'token': self.Token, 'sign': self.Sign, 'cuitRepresentada': self.Cuit},
            )
        ret = res['consultarDestinosCompraDJASReturn']
        return [("%(codigo)s\t%(descripcion)s"
                    % p['codigoDescripcion']).replace("\t", sep)
                 for p in ret['arrayCodigosDescripciones']]

    @inicializar_y_capturar_excepciones
    def ConsultarTiposReferencia(self, sep='|'):
        "Este m�todo retorna el conjunto de los tipos de referencia que pueden ser utilizados en la generaci�n de una solicitud de compra de divisas seg�n corresponda."
        res = self.client.consultarTiposReferencia(
            authRequest={'token': self.Token, 'sign': self.Sign, 'cuitRepresentada': self.Cuit},
            )
        ret = res['consultarTiposReferenciaReturn']
        return [("%(codigo)s\t%(descripcion)s"
                    % p['codigoDescripcion']).replace("\t", sep)
                 for p in ret['arrayCodigosDescripciones']]

    @inicializar_y_capturar_excepciones
    def ConsultarDestinosCompraTipoReferencia(self, tipo, sep='|'):
        "Este m�todo retorna el subconjunto de los destinos de compra de divisas alcanzados por algunas de las normativas vigentes seg�n el tipo de referencia requerido"
        res = self.client.consultarDestinosCompraTipoReferencia(
            authRequest={'token': self.Token, 'sign': self.Sign, 'cuitRepresentada': self.Cuit},
            tipo=tipo,
            )
        ret = res['consultarDestinosCompraTipoReferenciaReturn']
        return [("%(codigo)s\t%(descripcion)s"
                    % p['codigoDescripcion']).replace("\t", sep)
                 for p in ret['arrayCodigosDescripciones']]

    def LeerError(self):
        "Recorro los errores devueltos y devuelvo el primero si existe"

        if self.Errores:
            # extraigo el primer item
            er = self.Errores.pop(0)
            return er
        else:
            return ""

    def LeerErrorFormato(self):
        "Recorro los errores de formatos devueltos y devuelvo el primero si existe"

        if self.ErroresFormato:
            # extraigo el primer item
            er = self.ErroresFormato.pop(0)
            return er
        else:
            return ""

    def LeerInconsistencia(self):
        "Recorro las inconsistencias devueltas y devuelvo la primera si existe"

        if self.Inconsistencias:
            # extraigo el primer item
            er = self.Inconsistencias.pop(0)
            return er
        else:
            return ""

    def LoadTestXML(self, xml_file):
        class DummyHTTP:
            def __init__(self, xml_response):
                self.xml_response = xml_response
            def request(self, location, method, body, headers):
                return {}, self.xml_response
        self.client.http = DummyHTTP(open(xml_file).read())

    def AnalizarXml(self, xml=""):
        "Analiza un mensaje XML (por defecto la respuesta)"
        try:
            if not xml or xml=='XmlResponse':
                xml = self.XmlResponse
            elif xml=='XmlRequest':
                xml = self.XmlRequest
            self.xml = SimpleXMLElement(xml)
            return True
        except Exception as e:
            self.Excepcion = u"%s" % (e)
            return False

    def ObtenerTagXml(self, *tags):
        "Busca en el Xml analizado y devuelve el tag solicitado"
        # convierto el xml a un objeto
        try:
            if self.xml:
                xml = self.xml
                # por cada tag, lo busco segun su nombre o posici�n
                for tag in tags:
                    xml = xml(tag) # atajo a getitem y getattr
                # vuelvo a convertir a string el objeto xml encontrado
                return str(xml)
        except Exception as e:
            self.Excepcion = u"%s" % (e)


def p_assert_eq(a,b):
    print(a, a==b and '==' or '!=', b)


def main():
    "Funci�n principal de pruebas (obtener CAE)"
    import os, time

    DEBUG = '--debug' in sys.argv

    # obteniendo el TA
    TA = "TA-wscoc.xml"
    if not os.path.exists(TA) or os.path.getmtime(TA)+(60*60*5)<time.time():
        import wsaa
        tra = wsaa.create_tra(service="wscoc")
        cms = wsaa.sign_tra(tra,"olano.crt", "olanoycia.key")
        ta_string = wsaa.call_wsaa(cms, trace='--trace' in sys.argv)
        open(TA,"w").write(ta_string)

    # fin TA

    ws = WSCOC()
    ws.Cuit = "20267565393"

    ta_string=open(TA).read()
    ta = SimpleXMLElement(ta_string)
    ws.Token = str(ta.credentials.token)
    ws.Sign = str(ta.credentials.sign)

    ws.LanzarExcepciones = True
    ws.Conectar(wsdl=WSDL)

    if "--dummy" in sys.argv:
        try:
            ws.Dummy()
            print("AppServerStatus", ws.AppServerStatus)
            print("DbServerStatus", ws.DbServerStatus)
            print("AuthServerStatus", ws.AuthServerStatus)
        except Exception as e:
            raise
            print("Exception", e)
            print(ws.XmlRequest)
            print(ws.XmlResponse)

    if "--monedas" in sys.argv:
        print(ws.client.help("consultarMonedas"))
        try:
            for moneda in ws.ConsultarMonedas():
                print(moneda)
        except Exception as e:
            raise
            print("Exception", e)
            print(ws.XmlRequest)
            print(ws.XmlResponse)

    if "--motivos_ex_djai" in sys.argv:
        print(ws.ConsultarMotivosExcepcionDJAI())
    if "--destinos_djai" in sys.argv:
        print(ws.ConsultarDestinosCompraDJAI())

    if "--consultar_cuit" in sys.argv:
        print(ws.client.help("consultarCUIT"))
        try:
            print("Consultado CUITs....")
            nro_doc = 26756539
            tipo_doc = 96
            print(ws.ConsultarCUIT(nro_doc, tipo_doc))
            # recorro el detalle de los cuit devueltos:
            while ws.LeerCUITConsultado():
                print("CUIT", ws.CUITConsultada)
                print("Denominaci�n", ws.DenominacionConsultada)
        except Exception as e:
            raise
            print("Exception", e)
            print(ws.XmlRequest)
            print(ws.XmlResponse)

    if "--prueba" in sys.argv:
        print(ws.client.help("generarSolicitudCompraDivisa").encode("latin1"))
        try:
            cuit_comprador = 20267565393
            codigo_moneda = 1
            cotizacion_moneda = 4.26
            monto_pesos = 100
            cuit_representante = None
            codigo_destino = 625

            if "--loadxml" in sys.argv:
                ws.LoadTestXML("wscoc_response.xml")

            if not "--tur" in sys.argv:
                djai = "--djai" in sys.argv and "12345DJAI000001N" or None
                djas = "--djas" in sys.argv and "12001DJAS000901N" or None
                cod_ex_djai = "--no-djai" and 3 or None
                cod_ex_djas = "--no-djas" and 1 or None
                if '--ref' in sys.argv:
                    tipo, codigo = 1, '12345DJAI000067C'
                else:
                    tipo, codigo = None, None
                print("djai", djai)
                ok = ws.GenerarSolicitudCompraDivisa(cuit_comprador, codigo_moneda,
                                    cotizacion_moneda, monto_pesos,
                                    cuit_representante, codigo_destino,
                                    djai=djai, codigo_excepcion_djai=cod_ex_djai,
                                    djas=djas, codigo_excepcion_djas=cod_ex_djas,
                                    tipo=tipo, codigo=codigo,)
            else:
                print("Turista!")
                tipo_doc=91
                numero_doc=1234567
                apellido_nombre="Nombre y Apellido del turista extranjero"
                codigo_destino = 985
                ok = ws.GenerarSolicitudCompraDivisaTurExt(
                        tipo_doc, numero_doc, apellido_nombre,
                        codigo_moneda, cotizacion_moneda, monto_pesos,
                        cuit_representante, codigo_destino,
                        )
            while True:
                i = ws.LeerInconsistencia()
                if not i:
                    break
                print("Inconsistencia...", i)

            assert ok
            print('Resultado', ws.Resultado)
            assert ws.Resultado == 'A'
            print('COC', ws.COC)
            assert len(str(ws.COC)) == 12
            print("FechaEmisionCOC", ws.FechaEmisionCOC)
            print('CodigoSolicitud', ws.CodigoSolicitud)
            assert ws.CodigoSolicitud is not None
            print("EstadoSolicitud", ws.EstadoSolicitud)
            assert ws.EstadoSolicitud == 'OT'
            print("FechaEstado", ws.FechaEstado)
            print("DetalleCUITComprador", ws.CUITComprador, ws.DenominacionComprador)
            print("CodigoMoneda", ws.CodigoMoneda)
            assert ws.CodigoMoneda == 1
            print("CotizacionMoneda", ws.CotizacionMoneda)
            assert round(ws.CotizacionMoneda, 2) == 4.26
            print("MontoPesos", ws.MontoPesos)
            assert ws.MontoPesos - monto_pesos <= 0.01
            print("CodigoDestino", ws.CodigoDestino)
            assert ws.CodigoDestino == codigo_destino

            coc = ws.COC
            codigo_solicitud = ws.CodigoSolicitud
            # CO: confirmar, o 'DC' (desistio cliente) 'DB' (desistio banco)
            nuevo_estado =  'CO'
            ok = ws.InformarSolicitudCompraDivisa(codigo_solicitud, nuevo_estado)
            assert ok
            print('Resultado', ws.Resultado)
            assert ws.Resultado == 'A'
            print('COC', ws.COC)
            assert ws.COC == coc
            print("EstadoSolicitud", ws.EstadoSolicitud)
            assert ws.EstadoSolicitud == nuevo_estado

            ok = ws.AnularCOC(coc, cuit_comprador)
            assert ok
            print('Resultado', ws.Resultado)
            assert ws.Resultado == 'A'
            print('COC', ws.COC)
            assert ws.COC == coc
            print("EstadoSolicitud", ws.EstadoSolicitud)
            assert ws.EstadoSolicitud == 'AN'

            ok = ws.ConsultarSolicitudCompraDivisa(codigo_solicitud)
            assert ok
            print('CodigoSolicitud', ws.CodigoSolicitud)
            assert ws.CodigoSolicitud == codigo_solicitud
            print("EstadoSolicitud", ws.EstadoSolicitud)
            assert ws.EstadoSolicitud == 'AN'

        except:
            print(ws.XmlRequest)
            print(ws.XmlResponse)
            print(ws.ErrCode)
            print(ws.ErrMsg)
            raise

    if "--consultar_solicitudes" in sys.argv:
        cuit_comprador = None
        estado_solicitud = None
        fecha_emision_desde = '2011-11-01'
        fecha_emision_hasta = '2011-11-30'
        sols = ws.ConsultarSolicitudesCompraDivisas(cuit_comprador,
                                             estado_solicitud,
                                             fecha_emision_desde,
                                             fecha_emision_hasta,)
        # muestro los resultados de la b�squeda
        print("Solicitudes consultadas:")
        for sol in sols:
            print("C�digo de Solicitud:", sol)
            # podr�a llamar a ws.ConsultarSolicitudCompraDivisa
        print("hecho.")

        ws.AnalizarXml("XmlResponse")

        # recorro las solicitudes devueltas
        i = 0
        while ws.LeerSolicitudConsultada():
            print("-" * 80)
            coc = ws.ObtenerTagXml('arrayDetallesSolicitudes', 'detalleSolicitudes', i, 'coc')
            cuit = ws.ObtenerTagXml('arrayDetallesSolicitudes', 'detalleSolicitudes', i, 'cuit')
            p_assert_eq(coc, ws.COC)
            print('CUIT', cuit)
            print("FechaEmisionCOC", ws.FechaEmisionCOC)
            print('CodigoSolicitud', ws.CodigoSolicitud)
            print("EstadoSolicitud", ws.EstadoSolicitud)
            print("FechaEstado", ws.FechaEstado)
            print("DetalleCUITComprador", ws.CUITComprador, ws.DenominacionComprador)
            print("CodigoMoneda", ws.CodigoMoneda)
            print("CotizacionMoneda", ws.CotizacionMoneda)
            print("MontoPesos", ws.MontoPesos)
            print("CodigoDestino", ws.CodigoDestino)
            print("=" * 80)
            i = i + 1



    if "--parametros" in sys.argv:
        print("=== Tipos de Estado Solicitud ===")
        print(u'\n'.join(["||%s||" % s for s in ws.ConsultarTiposEstadoSolicitud(sep="||")]))
        print("=== Monedas ===")
        print(u'\n'.join(["||%s||" % s for s in ws.ConsultarMonedas(sep="||")]))
        print("=== Destinos de Compra ===")
        print(u'\n'.join(["||%s||" % s for s in ws.ConsultarDestinosCompra(sep="||")]))
        print("=== Tipos de Documento ===")
        print(u'\n'.join(["||%s||" % s for s in ws.ConsultarTiposDocumento(sep="||")]))
        print("=== Tipos Estado Solicitud ===")
        print(u'\n'.join(["||%s||" % s for s in ws.ConsultarTiposEstadoSolicitud(sep="||")]))
        print("=== Motivos Excepcion DJAI ===")
        print(u'\n'.join(["||%s||" % s for s in ws.ConsultarMotivosExcepcionDJAI(sep="||")]))
        print("=== Destinos Compra DJAI ===")
        print(u'\n'.join(["||%s||" % s for s in ws.ConsultarDestinosCompraDJAI(sep="||")]))
        print("=== Motivos Excepcion DJAS ===")
        print(u'\n'.join(["||%s||" % s for s in ws.ConsultarMotivosExcepcionDJAS(sep="||")]))
        print("=== Destinos Compra DJAS ===")
        print(u'\n'.join(["||%s||" % s for s in ws.ConsultarDestinosCompraDJAS(sep="||")]))

    if "--consultar_djai" in sys.argv:
        djai = "12345DJAI000001N"
        cuit = 20267565393
        ws.ConsultarDJAI(djai, cuit)
        print("DJAI", ws.DJAI)
        print("Monto FOB", ws.MontoFOB)
        print("Codigo Moneda", ws.CodigoMoneda)
        print("Estado DJAI", ws.EstadoDJAI)
        print("Errores", ws.Errores)
        print("ErroresFormato", ws.ErroresFormato)

    if "--consultar_djas" in sys.argv:
        djas = "12001DJAS000901N"
        cuit = 20267565393
        ws.ConsultarDJAS(djas, cuit)
        print("DJAS", ws.DJAS)
        print("Monto FOB", ws.MontoFOB)
        print("Codigo Moneda", ws.CodigoMoneda)
        print("Estado DJAI", ws.EstadoDJAI)
        print("Errores", ws.Errores)
        print("ErroresFormato", ws.ErroresFormato)

    if "--consultar_ref" in sys.argv:
        codigo = "12345DJAI000067C"
        cuit = 20267565393
        ws.ConsultarReferencia(1, codigo)
        print("Monto FOB", ws.MontoFOB)
        print("Codigo Moneda", ws.CodigoMoneda)
        print("Estado", ws.Estado)
        print("Errores", ws.Errores)
        print("ErroresFormato", ws.ErroresFormato)

    if "--consultar_dest_ref" in sys.argv:
        print(ws.ConsultarDestinosCompraTipoReferencia(1))

    if "--consultar_tipos_ref" in sys.argv:
        print(ws.ConsultarTiposReferencia())

# busco el directorio de instalaci�n (global para que no cambie si usan otra dll)
if not hasattr(sys, "frozen"):
    basepath = __file__
elif sys.frozen=='dll':
    import win32api
    basepath = win32api.GetModuleFileName(sys.frozendllhandle)
else:
    basepath = sys.executable
INSTALL_DIR = os.path.dirname(os.path.abspath(basepath))

if __name__ == '__main__':

    if "--register" in sys.argv or "--unregister" in sys.argv:
        import win32com.server.register
        win32com.server.register.UseCommandLine(WSCOC)
    else:
        main()



#!/usr/bin/python
# -*- coding: latin-1 -*-
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by the
# Free Software Foundation; version 3.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTIBILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.

"Módulo con funciones auxiliares para el manejo de errores y temas comunes"

__author__ = "Mariano Reingart <reingart@gmail.com>"
__copyright__ = "Copyright (C) 2013 Mariano Reingart"
__license__ = "GPL 3.0"

import datetime
import functools
import locale
import socket
import sys
import os
# import stat
import traceback
import warnings
from io import StringIO
from decimal import Decimal
from urllib.parse import urlencode
import mimetypes
from email.generator import _make_boundary
from html.parser import HTMLParser
from http.cookies import SimpleCookie
from configparser import SafeConfigParser

from py3simplesoap.client import (
    SimpleXMLElement,
    SoapClient,
    SoapFault,
    parse_proxy,
    set_http_wrapper
)


try:
    import dbf
except ImportError:
    print("para soporte de DBF debe instalar dbf 0.88.019 o superior")

try:
    import json
except ImportError:
    try:
        import simplejson as json
    except:
        print("para soporte de JSON debe instalar simplejson")
        json = None

try:
    import httplib2
    # corregir temas de negociacion de SSL en algunas versiones de ubuntu:
    import platform
    dist, ver, nick = platform.linux_distribution() if sys.version > (
        2, 6
    ) else ("", "", "")
    # from py3simplesoap.client import SoapClient
    monkey_patch = httplib2._ssl_wrap_socket.__module__ != "httplib2"
    if dist == 'Ubuntu' and ver == '14.04' and not monkey_patch:
        import ssl

        def _ssl_wrap_socket(sock, key_file, cert_file,
                             disable_validation, ca_certs):
            if disable_validation:
                cert_reqs = ssl.CERT_NONE
            else:
                cert_reqs = ssl.CERT_REQUIRED
            return ssl.wrap_socket(
                sock, keyfile=key_file, certfile=cert_file,
                cert_reqs=cert_reqs, ca_certs=ca_certs,
                ssl_version=ssl.PROTOCOL_SSLv3
            )
        httplib2._ssl_wrap_socket = _ssl_wrap_socket

except:
    print("para soporte de WebClient debe instalar httplib2")


DEBUG = False

import logging
_logger = logging.getLogger(__name__)
# Funciones para manejo de errores:


def exception_info(current_filename=None, index=-1):
    "Analizar el traceback y armar un dict con la info amigable user-friendly"
    # guardo el traceback original (por si hay una excepción):
    info = sys.exc_info()  # exc_type, exc_value, exc_traceback
    # importante: no usar unpacking porque puede causar memory leak
    if not current_filename:
        # genero un call stack para ver quien me llamó y limitar la traza:
        # advertencia: esto es necesario ya que en py2exe no tengo __file__
        try:
            raise ZeroDivisionError
        except ZeroDivisionError:
            f = sys.exc_info()[2].tb_frame.f_back
        current_filename = os.path.normpath(
            os.path.abspath(f.f_code.co_filename)
        )

    # extraer la última traza del archivo solicitado:
    # (útil para no alargar demasiado la traza con lineas de las librerías)
    ret = {'filename': "", 'lineno': 0, 'function_name': "", 'code': ""}
    try:
        for (filename, lineno, fn, text) in traceback.extract_tb(info[2]):
            if os.path.normpath(os.path.abspath(filename)) == current_filename:
                ret = {'filename': filename, 'lineno': lineno,
                       'function_name': fn, 'code': text}
    except Exception as e:
        print(str(e))
        pass
    # obtengo el mensaje de excepcion tal cual lo formatea python:
    # (para evitar errores de encoding)
    try:
        ret['msg'] = traceback.format_exception_only(*info[0:2])[0]
    except:
        ret['msg'] = '<no disponible>'
    # obtener el nombre de la excepcion (ej. "NameError")
    try:
        ret['name'] = info[0].__name__
    except:
        ret['name'] = 'Exception'
    # obtener la traza formateada como string:
    try:
        tb = traceback.format_exception(*info)
        ret['tb'] = ''.join(tb)
    except:
        ret['tb'] = ""
    return ret


def json_serializer(obj):
    if isinstance(obj, (datetime.datetime, datetime.date)):
        return obj.isoformat()
    else:
        return str(obj)


def inicializar_y_capturar_excepciones(func):
    "Decorador para inicializar y capturar errores (version para webservices)"
    @functools.wraps(func)
    def capturar_errores_wrapper(self, *args, **kwargs):
        try:
            # inicializo (limpio variables)
            self.Errores = []           # listas de str para lenguajes legados
            self.Observaciones = []
            self.errores = []           # listas de dict para usar en python
            self.observaciones = []
            self.Eventos = []
            self.Traceback = self.Excepcion = ""
            self.ErrCode = self.ErrMsg = self.Obs = ""
            # limpio variables especificas del webservice:
            self.inicializar()
            # actualizo los parámetros
            kwargs.update(self.params_in)
            # limpio los parámetros
            self.params_in = {}
            self.params_out = {}
            # llamo a la función (con reintentos)
            retry = self.reintentos + 1
            while retry:
                try:
                    retry -= 1
                    return func(self, *args, **kwargs)
                except socket.error as e:
                    if e.errno not in (10054, 10053):
                        # solo reintentar si el error es de conexión
                        # (10054, 'Connection reset by peer')
                        # (10053, 'Software caused connection abort')
                        raise
                    else:
                        if DEBUG:
                            print((e, "Reintentando..."))
                        self.log(exception_info().get("msg", ""))

        except SoapFault as e:
            # guardo destalle de la excepción SOAP
            self.ErrCode = str(e.faultcode)
            self.ErrMsg = str(e.faultstring)
            self.Excepcion = "%s: %s" % (e.faultcode, e.faultstring, )
            if self.LanzarExcepciones:
                raise
        except Exception as e:
            ex = exception_info()
            self.Traceback = ex.get("tb", "")
            try:
                self.Excepcion = ex.get("msg", "")
            except:
                self.Excepcion = "<no disponible>"
            if self.LanzarExcepciones:
                raise
            else:
                self.ErrMsg = self.Excepcion
        finally:
            # guardo datos de depuración
            if self.client:
                self.XmlRequest = self.client.xml_request
                self.XmlResponse = self.client.xml_response
    return capturar_errores_wrapper


def inicializar_y_capturar_excepciones_simple(func):
    "Decorador para inicializar y capturar errores (versión básica indep.)"
    @functools.wraps(func)
    def capturar_errores_wrapper(self, *args, **kwargs):
        self.inicializar()
        try:
            return func(self, *args, **kwargs)
        except:
            ex = exception_info()
            self.Excepcion = ex['name']
            self.Traceback = ex['msg']
            if self.LanzarExcepciones:
                raise
            else:
                return False
    return capturar_errores_wrapper


class BaseWS:
    "Infraestructura basica para interfaces webservices de AFIP"

    def __init__(self, reintentos=1):
        self.reintentos = reintentos
        self.xml = self.client = self.Log = None
        self.params_in = {}
        self.inicializar()
        self.Token = self.Sign = ""
        self.LanzarExcepciones = True

    def inicializar(self):
        self.Excepcion = self.Traceback = ""
        self.XmlRequest = self.XmlResponse = ""

    def Conectar(
        self, cache=None, wsdl=None, proxy="", wrapper=None, cacert=None,
        timeout=30, soap_server=None
    ):
        "Conectar cliente soap del web service"
        try:
            self.Excepcion = "Begin try"
            # analizar transporte y servidor proxy:
            if wrapper:
                Http = set_http_wrapper(wrapper)
                self.Version = self.Version + " " + Http._wrapper_version
            self.Excepcion = "after wrapper"
            if isinstance(proxy, dict):
                proxy_dict = proxy
            else:
                proxy_dict = parse_proxy(proxy)
                self.log("Proxy Dict: %s" % str(proxy_dict))
            self.Excepcion = "after proxy_dict"
            if self.HOMO or not wsdl:
                wsdl = self.WSDL
            self.Excepcion = "after self.HOMO or not wsdl"
            # agregar sufijo para descargar descripción del servicio ?WSDL o
            # ?wsdl
            if not wsdl.endswith(self.WSDL[-5:]) and wsdl.startswith("http"):
                wsdl += self.WSDL[-5:]
            self.Excepcion = "after not wsdl.endswith"
            if not cache or self.HOMO:
                # use 'cache' from installation base directory
                cache = os.path.join(self.InstallDir, 'cache')
            self.Excepcion = "after self.HOMO or not cache"
            # deshabilitar verificación cert. servidor si es nulo falso vacio
            if not cacert:
                cacert = None
            elif cacert is True:
                # usar certificados predeterminados que vienen en la biblioteca
                cacert = os.path.join(httplib2.__path__[0], 'cacerts.txt')
                self.Excepcion = " elif cacert is True: "
            elif cacert.startswith("-----BEGIN CERTIFICATE-----"):
                pass
            else:
                if not os.path.exists(cacert):
                    self.log("Buscando CACERT en conf...")
                    cacert = os.path.join(
                        self.InstallDir, "conf", os.path.basename(cacert)
                    )
                if cacert and not os.path.exists(cacert):
                    self.log("No se encuentra CACERT: %s" % str(cacert))
                    warnings.warn("No se encuentra CACERT: %s" % str(cacert))
                    cacert = None   # wrong version, certificates not found...
                    self.Excepcion = "No se encuentra CACERT"
                    raise RuntimeError(
                        "Error de configuracion CACERT ver DebugLog"
                    )
                    return False
                self.Excepcion = "else"

            _logger.warn('____________________A_______________________ ')

            self.Excepcion = "Conectando a wsdl=%s cache=%s proxy=%s" % (
                wsdl, cache, proxy_dict
            )
            self.log(
                "Conectando a wsdl=%s cache=%s proxy=%s" % (
                    wsdl, cache, proxy_dict
                )
            )
            # analizar espacio de nombres (axis vs .net):
            ns = 'ser' if self.WSDL[-5:] == "?wsdl" else None
            self.Excepcion = str(soap_server)
            _logger.warn('_____________________B______________________ ')

            self.client = SoapClient(
                wsdl=wsdl,
                cache=cache,
                proxy=proxy_dict,
                cacert=cacert,
                timeout=timeout,
                ns=ns,
                soap_server=soap_server,
                trace="--trace" in sys.argv)

            _logger.warn('____________________C_______________________ ')
            self.cache = cache  # utilizado por WSLPG y WSAA (Ticket de Acceso)
            _logger.warn('____________________D_______________________ ')
            # utilizado por TrazaMed (para corregir el location)
            self.wsdl = wsdl
            _logger.warn('____________________E________________________ ')
            # corrijo ubicación del servidor (puerto http 80 en el WSDL AFIP)
            for service in list(self.client.services.values()):
                for port in list(service['ports'].values()):
                    location = port['location']
                    if location and location.startswith("http://"):
                        warnings.warn("Corrigiendo WSDL ... %s" % location)
                        location = location.replace("http://", "https://")\
                            .replace(":80", ":443")
                        port['location'] = location
            return True
        except:
            if self.LanzarExcepciones:
                raise
            return False

    def log(self, msg):
        "Dejar mensaje en bitacora de depuración (método interno)"
        if not isinstance(msg, str):
            msg = str(msg, 'utf8', 'ignore')
        if not self.Log:
            self.Log = StringIO()
        self.Log.write(msg)
        self.Log.write('\n\r')
        if DEBUG:
            warnings.warn(msg)

    def DebugLog(self):
        "Devolver y limpiar la bitácora de depuración"
        if self.Log:
            msg = self.Log.getvalue()
            # limpiar log
            self.Log.close()
            self.Log = None
        else:
            msg = ''
        return msg

    def LoadTestXML(self, xml):
        "Cargar un archivo de pruebas con la respuesta simulada (depuración)"
        # si el parametro es un nombre de archivo, cargar el contenido:
        if os.path.exists(xml):
            xml = open(xml).read()

        class DummyHTTP:
            def __init__(self, xml_response):
                self.xml_response = xml_response

            def request(self, location, method, body, headers):
                return {}, self.xml_response
        self.client.http = DummyHTTP(xml)

    @property
    def xml_request(self):
        return self.XmlRequest

    @property
    def xml_response(self):
        return self.XmlResponse

    def AnalizarXml(self, xml=""):
        "Analiza un mensaje XML (por defecto el ticket de acceso)"
        try:
            if not xml or xml == 'XmlResponse':
                xml = self.XmlResponse
            elif xml == 'XmlRequest':
                xml = self.XmlRequest
            self.xml = SimpleXMLElement(xml)
            return True
        except Exception as e:
            print(str(e))
            self.Excepcion = traceback.format_exception_only(
                sys.exc_info()[0], sys.exc_info()[1]
            )[0]
            return False

    def ObtenerTagXml(self, *tags):
        "Busca en el Xml analizado y devuelve el tag solicitado"
        # convierto el xml a un objeto
        try:
            if self.xml:
                xml = self.xml
                # por cada tag, lo busco segun su nombre o posición
                for tag in tags:
                    xml = xml(tag)  # atajo a getitem y getattr
                # vuelvo a convertir a string el objeto xml encontrado
                return str(xml)
        except Exception as e:
            print(str(e))
            self.Excepcion = traceback.format_exception_only(
                sys.exc_info()[0], sys.exc_info()[1]
            )[0]

    def SetParametros(self, cuit, token, sign):
        "Establece un parámetro general"
        self.Token = token
        self.Sign = sign
        self.Cuit = cuit
        return True

    @inicializar_y_capturar_excepciones
    def SetTicketAcceso(self, ta_string):
        "Establecer el token y sign desde un ticket de acceso XML"
        if ta_string:
            ta = SimpleXMLElement(ta_string)
            self.Token = str(ta.credentials.token)
            self.Sign = str(ta.credentials.sign)
            return True
        else:
            raise RuntimeError("Ticket de Acceso vacio!")

    def SetParametro(self, clave, valor):
        "Establece un parámetro de entrada (a usarse en llamada posterior)"
        # útil para parámetros de entrada (por ej. VFP9 no soporta más de 27)
        self.params_in[str(clave)] = valor
        return True

    def GetParametro(
        self, clave, clave1=None, clave2=None, clave3=None, clave4=None
    ):
        "Devuelve un parámetro de salida (establecido por llamada anterior)"
        # útil para parámetros de salida (por ej. campos de TransaccionPlainWS)
        valor = self.params_out.get(clave)
        # busco datos "anidados" (listas / diccionarios)
        for clave in (clave1, clave2, clave3, clave4):
            if clave is not None and valor is not None:
                if isinstance(clave1, str) and clave.isdigit():
                    clave = int(clave)
                try:
                    valor = valor[clave]
                except (KeyError, IndexError):
                    valor = None
        if valor is not None:
            if isinstance(valor, str):
                return valor
            else:
                return str(valor)
        else:
            return ""

    def LeerError(self):
        "Recorro los errores devueltos y devuelvo el primero si existe"

        if self.Errores:
            # extraigo el primer item
            er = self.Errores.pop(0)
            return er
        else:
            return ""


class WebClient:
    """Minimal webservice client to do POST request with multipart encoded
    FORM data"""

    def __init__(
        self, location, enctype="multipart/form-data", trace=False,
        cacert=None, timeout=30
    ):
        kwargs = {}
        if httplib2.__version__ >= '0.3.0':
                kwargs['timeout'] = timeout
        if httplib2.__version__ >= '0.7.0':
                kwargs['disable_ssl_certificate_validation'] = cacert is None
                kwargs['ca_certs'] = cacert
        self.http = httplib2.Http(**kwargs)
        self.trace = trace
        self.location = location
        self.enctype = enctype
        self.cookies = None
        self.method = "POST"
        self.referer = None

    def multipart_encode(self, vars):
        "Enconde form data (vars dict)"
        boundary = _make_boundary()
        buf = StringIO()
        for key, value in list(vars.items()):
            if not isinstance(value, file):
                buf.write('--%s\r\n' % boundary)
                buf.write('Content-Disposition: form-data; name="%s"' % key)
                buf.write('\r\n\r\n' + value + '\r\n')
            else:
                fd = value
                # file_size = os.fstat(fd.fileno())[stat.ST_SIZE]
                filename = os.path.basename(fd.name)
                contenttype = mimetypes.guess_type(filename)[0] or \
                    'application/octet-stream'
                buf.write('--%s\r\n' % boundary)
                buf.write('Content-Disposition: form-data; name="%s"; \
                    filename="%s"\r\n' % (key, filename))
                buf.write('Content-Type: %s\r\n' % contenttype)
                # buffer += 'Content-Length: %s\r\n' % file_size
                fd.seek(0)
                buf.write('\r\n' + fd.read() + '\r\n')
        buf.write('--' + boundary + '--\r\n\r\n')
        buf = buf.getvalue()
        return boundary, buf

    def __call__(self, *args, **vars):
        "Perform a GET/POST request and return the response"
        location = self.location
        # if isinstance(location, str):
        #     location = location.encode("utf8")

        # extend the base URI with additional components
        if args:
            _logger.warn('args')
            _logger.warn(args)
            location += "/".join(args)
        if self.method == "GET":
            location += "?%s" % urlencode(vars)

        # prepare the request content suitable to be sent to the server:
        if self.enctype == "multipart/form-data":
            boundary, body = self.multipart_encode(vars)
            content_type = '%s; boundary=%s' % (self.enctype, boundary)
        elif self.enctype == "application/x-www-form-urlencoded":
            body = urlencode(vars)
            content_type = self.enctype
        else:
            body = None

        # add headers according method, cookies, etc.:
        headers = {}
        if self.method == "POST":
            headers.update({
                'Content-type': content_type,
                'Content-length': str(len(body)),
                })
        if self.cookies:
            headers['Cookie'] = self.cookies.output(
                attrs=(), header="", sep=";"
            )
        if self.referer:
            headers['Referer'] = self.referer

        if self.trace:
            print(("-"*80))
            print(("%s %s" % (self.method, location)))
            print(('\n'.join(["%s: %s" % (k, v) for k, v in list(
                headers.items()
            )])))
            print(("\n%s" % body))

        # send the request to the server and store the result:
        response, content = self.http.request(
            location, self.method, body=body, headers=headers)
        self.response = response
        self.content = content

        if self.trace:
            print()
            print(
                ('\n'.join(
                    ["%s: %s" % (k, v) for k, v in list(response.items())]
                ))
            )
            print(content)
            print(("="*80))

        # Parse and store the cookies (if any)
        if "set-cookie" in self.response:
            if not self.cookies:
                self.cookies = SimpleCookie()
            self.cookies.load(self.response["set-cookie"])

        return content


class AttrDict(dict):
    "Custom Dict to hold attributes and items"


class HTMLFormParser(HTMLParser):
    "Convert HTML form into custom named-tuple dicts"

    def __init__(self, *args, **kwargs):
        HTMLParser.__init__(self, *args, **kwargs)
        self.forms = {}

    def handle_starttag(self, tag, attrs):
        attrs = dict(attrs)
        if 'name' in attrs:
            name = attrs['name']
        elif 'id' in attrs:
            name = attrs['id']
        else:
            name = None
        if tag == 'form':
            form = AttrDict()
            for k, v in list(attrs.items()):
                setattr(form, "_%s" % k, v)
            self.form = self.forms[name or len(self.forms)] = form
        elif tag == 'input':
            self.form[name or len(self.form)] = attrs.get('value')


# Funciones para manejo de archivos de texto de campos de ancho fijo:


def leer(linea, formato, expandir_fechas=False):
    "Analiza una linea de texto dado un formato, devuelve un diccionario"
    dic = {}
    comienzo = 1
    for fmt in formato:
        clave, longitud, tipo = fmt[0:3]
        dec = (len(fmt) > 3 and isinstance(fmt[3], int)) and fmt[3] or 2
        valor = linea[comienzo-1:comienzo-1+longitud].strip()
        try:
            if chr(8) in valor or chr(127) in valor or chr(255) in valor:
                valor = None        # nulo
            elif tipo == N:
                if valor:
                    valor = int(valor)
                else:
                    valor = 0
            elif tipo == I:
                if valor:
                    try:
                        if '.' in valor:
                                valor = float(valor)
                        else:
                            valor = valor.strip(" ")
                            valor = float(
                                ("%%s.%%0%sd" % dec) % (
                                    int(valor[:-dec] or '0'),
                                    int(valor[-dec:] or '0')
                                )
                            )
                    except ValueError:
                        raise ValueError("Campo invalido: %s = '%s'" % (
                            clave, valor
                        ))
                else:
                    valor = 0.00
            elif expandir_fechas and clave.lower().startswith("fec") and \
                    longitud <= 8:
                if valor:
                    valor = "%s-%s-%s" % (valor[0:4], valor[4:6], valor[6:8])
                else:
                    valor = None
            else:
                valor = valor.decode("ascii", "ignore")
            dic[clave] = valor
            comienzo += longitud
        except Exception as e:
            raise ValueError("Error al leer campo %s pos %s val '%s': %s" % (
                clave, comienzo, valor, str(e)))
    return dic


def escribir(dic, formato, contraer_fechas=False):
    "Genera una cadena dado un formato y un diccionario de claves/valores"
    linea = " " * sum([fmt[1] for fmt in formato])
    comienzo = 1
    for fmt in formato:
        clave, longitud, tipo = fmt[0:3]
        try:
            dec = (len(fmt) > 3 and isinstance(fmt[3], int)) and fmt[3] or 2
            if clave.capitalize() in dic:
                clave = clave.capitalize()
            s = dic.get(clave, "")
            if isinstance(s, str):
                s = s.encode("latin1")
            if s is None:
                valor = ""
            else:
                valor = str(s)
            # reemplazo saltos de linea por tabulaci{on vertical
            valor = valor.replace("\n\r", "\v").replace("\n", "\v")\
                .replace("\r", "\v")
            if tipo == N and valor and valor != "NULL":
                valor = ("%%0%dd" % longitud) % int(valor)
            elif tipo == I and valor:
                valor = ("%%0%d.%df" % (longitud+1, dec) % float(valor))\
                    .replace(".", "")
            elif contraer_fechas and clave.lower().startswith("fec") and \
                    longitud <= 8 and valor:
                valor = valor.replace("-", "")
            else:
                valor = ("%%-0%ds" % longitud) % valor
            linea = linea[:comienzo-1] + valor + linea[comienzo-1+longitud:]
            comienzo += longitud
        except Exception as e:
            raise ValueError(
                "Error al escribir campo %s pos %s val '%s': %s" % (
                    clave, comienzo, valor, str(e)
                )
            )
    return linea + "\n"


# Tipos de datos (código RG1361)


N = 'Numerico'      # 2
A = 'Alfanumerico'  # 3
I = 'Importe'       # 4
C = A               # 1 (caracter alfabetico)
B = A               # 9 (blanco)


# Funciones para manejo de tablas en DBF


def guardar_dbf(formatos, agrega=False, conf_dbf=None):
    # import dbf
    if DEBUG:
        print("Creando DBF...")

    # tablas = {}
    for nombre, formato, l in formatos:
        campos = []
        claves = []
        filename = conf_dbf.get(nombre.lower(), "%s.dbf" % nombre[:8])
        if DEBUG:
            print(("=== tabla %s (%s) ===" % (nombre, filename)))
        for fmt in formato:
            clave, longitud, tipo = fmt[0:3]
            dec = len(fmt) > 3 and fmt[3] or (tipo == 'I' and '2' or '')
            if longitud > 250:
                tipo = "M"  # memo!
            elif tipo == A:
                tipo = "C(%s)" % longitud
            elif tipo == N:
                if longitud >= 18:
                    longitud = 17
                tipo = "N(%s,0)" % longitud
            elif tipo == I:
                if longitud >= 18:
                    longitud = 17
                if longitud - 2 <= dec:
                    longitud += longitud - dec + 1  # ajusto long. decimales
                tipo = "N(%s,%s)" % (longitud, dec)
            clave_dbf = dar_nombre_campo_dbf(clave, claves)
            campo = "%s %s" % (clave_dbf, tipo)
            if DEBUG:
                print((" * %s : %s" % (campo, clave)))
            campos.append(campo)
            claves.append(clave_dbf)
        if DEBUG:
            print(("leyendo tabla", nombre, filename))
        if agrega:
            tabla = dbf.Table(filename, campos)
        else:
            tabla = dbf.Table(filename)

        for d in l:
            # si no es un diccionario, ignorar ya que seguramente va en otra
            # tabla (por ej. retenciones tiene su propio formato)
            if isinstance(d, str):
                continue
            r = {}
            claves = []
            for fmt in formato:
                clave, longitud, tipo = fmt[0:3]
                if agrega or clave in d:
                    v = d.get(clave, None)
                    if DEBUG:
                        print((clave, v, tipo))
                    if v is None and tipo == A:
                        v = ''
                    if (v is None or v == '') and tipo in (I, N):
                        v = 0
                    if tipo == A:
                        if isinstance(v, str):
                            v = v.encode("ascii", "replace")
                        if isinstance(v, str):
                            v = v.decode("ascii", "replace")\
                                .encode("ascii", "replace")
                        if not isinstance(v, str):
                            v = str(v)
                        if len(v) > longitud:
                            # recorto el string para que quepa
                            v = v[:longitud]
                    clave_dbf = dar_nombre_campo_dbf(clave, claves)
                    claves.append(clave_dbf)
                    r[clave_dbf] = v
            # agregar si lo solicitaron o si la tabla no tiene registros:
            if agrega or not tabla:
                if DEBUG:
                    print(("Agregando !!!", r))
                registro = tabla.append(r)
                print('registro: %s' % str(registro))
            else:
                if DEBUG:
                    print(("Actualizando ", r))
                reg = tabla.current()
                for k, v in list(reg.scatter_fields().items()):
                    if k not in r:
                        r[k] = v
                if DEBUG:
                    print(("Actualizando ", r))
                reg.write_record(**r)
                # mover de registro para no actualizar siempre el primero:
                if not tabla.eof() and len(l) > 1:
                    if DEBUG:
                        print((
                            "Moviendo al próximo registro ",
                            tabla.record_number
                        ))
                    next(tabla)
        tabla.close()


def leer_dbf(formatos, conf_dbf):
    import dbf
    if DEBUG:
        print("Leyendo DBF...")

    for nombre, formato, ld in formatos:
        filename = conf_dbf.get(nombre.lower(), "%s.dbf" % nombre[:8])
        if DEBUG:
            print(("leyendo tabla", nombre, filename))
        if not os.path.exists(filename):
            continue
        tabla = dbf.Table(filename)
        for reg in tabla:
            r = {}
            d = reg.scatter_fields()
            claves = []
            for fmt in formato:
                clave, longitud, tipo = fmt[0:3]
                #import pdb; pdb.set_trace()
                clave_dbf = dar_nombre_campo_dbf(clave, claves)
                claves.append(clave_dbf)
                v = d.get(clave_dbf)
                r[clave] = v
            if isinstance(ld, dict):
                ld.update(r)
            else:
                ld.append(r)


def dar_nombre_campo_dbf(clave, claves):
    "Reducir nombre de campo a 10 caracteres, sin espacios ni _, sin repetir"
    # achico el nombre del campo para que quepa en la tabla:
    nombre = clave.replace("_", "")[:10]
    # si el campo esta repetido, le agrego un número
    i = 0
    while nombre in claves:
        i += 1
        nombre = nombre[:9] + str(i)
    return nombre.lower()


def verifica(ver_list, res_dict, difs):
    "Verificar que dos diccionarios sean iguales, actualiza lista diferencias"
    for k, v in list(ver_list.items()):
        # normalizo a float para poder comparar numericamente:
        if isinstance(v, (Decimal, int)):
            v = float(v)
        if isinstance(res_dict.get(k), (Decimal, int)):
            res_dict[k] = float(res_dict[k])
        if isinstance(v, list):
            # verifico que ambas listas tengan la misma cantidad de elementos:
            if v and not k in res_dict and v:
                difs.append("falta tag %s: %s %s" % (
                    k, repr(v), repr(res_dict.get(k))
                ))
            elif len(res_dict.get(k, [])) != len(v or []):
                difs.append(
                    "tag %s len !=: %s %s" % (
                        k, repr(v), repr(res_dict.get(k))
                    )
                )
            else:
                # ordeno las listas para poder compararlas si vienen mezcladas
                rl = sorted(res_dict.get(k, []))
                # comparo los elementos uno a uno:
                for i, vl in enumerate(sorted(v)):
                    verifica(vl, rl[i], difs)
        elif isinstance(v, dict):
            # comparo recursivamente los elementos:
            verifica(v, res_dict.get(k, {}), difs)
        elif res_dict.get(k) is None or v is None:
            # alguno de los dos es nulo, verifico si ambos lo son o faltan
            if v == "":
                v = None
            r = res_dict.get(k)
            if r == "":
                r = None
            if not (r is None and v is None):
                difs.append("%s: nil %s!=%s" % (k, repr(v), repr(r)))
        elif type(res_dict.get(k)) == type(v):
            # tipos iguales, los comparo directamente
            if res_dict.get(k) != v:
                difs.append("%s: %s!=%s" % (k, repr(v), repr(res_dict.get(k))))
        elif isinstance(v, float) or isinstance(res_dict.get(k), float):
            # comparar numericamente
            if float(res_dict.get(k)) != float(v):
                difs.append("%s: %s!=%s" % (k, repr(v), repr(res_dict.get(k))))
        elif str(res_dict.get(k)) != str(v):
            # tipos diferentes, comparo la representación
            difs.append("%s: str %s!=%s" % (k, repr(v), repr(res_dict.get(k))))
        else:
            pass
            # print("%s: %s==%s" % (k, repr(v), repr(res_dict[k])))


def safe_console():
    if False and sys.stdout.encoding is None:
        class SafeWriter:
            def __init__(self, target):
                self.target = target
                self.encoding = 'utf-8'
                self.errors = 'replace'
                self.encode_to = 'latin-1'

            def write(self, s):
                self.target.write(self.intercept(s))

            def flush(self):
                self.target.flush()

            def intercept(self, s):
                if not isinstance(s, str):
                    s = s.decode(self.encode_to, self.errors)
                return s.encode(self.encoding, self.errors)

        sys.stdout = SafeWriter(sys.stdout)
        #sys.stderr = SafeWriter(sys.stderr)
        print(("Encodign in %s" % locale.getpreferredencoding()))


def abrir_conf(config_file, debug=False):
    "Abrir el archivo de configuración (usar primer parámetro como ruta)"
    # en principio, usar el nombre de archivo predeterminado
    # si se pasa el archivo de configuración por parámetro, confirmar que
    # exista
    # y descartar que sea una opción
    if len(sys.argv) > 1:
        if os.path.splitext(sys.argv[1])[1].lower() == ".ini":
            config_file = sys.argv.pop(1)
    if not os.path.exists(config_file) or not os.path.isfile(config_file):
        warnings.warn("Archivo de configuracion %s invalido" % config_file)

    if debug:
        print("CONFIG_FILE:", config_file)

    config = SafeConfigParser()
    config.read(config_file, encoding="latin1")

    return config


def get_install_dir():
    if not hasattr(sys, "frozen"):
        basepath = __file__
    elif sys.frozen == 'dll':
        import win32api
        basepath = win32api.GetModuleFileName(sys.frozendllhandle)
    else:
        basepath = sys.executable

    if hasattr(sys, "frozen"):
        # we are running as py2exe-packed executable
        import pythoncom
        pythoncom.frozen = 1
        sys.argv[0] = sys.executable

    return os.path.dirname(os.path.abspath(basepath))


if __name__ == "__main__":
    print((get_install_dir()))
    try:
        1/0
    except:
        ex = exception_info()
        print(ex)
        assert ex['name'] == "ZeroDivisionError"
        assert ex['lineno'] == 73
        assert ex['tb']
